# Copyright 2024 Marimo. All rights reserved.
from urllib.parse import urlparse
from xml.etree.ElementTree import Element

from markdown import Extension, Markdown, treeprocessors  # type: ignore

# Adapted from https://github.com/squidfunk/mkdocs-material/discussions/3660#discussioncomment-6725823  # noqa: E501


class ExternalLinksTreeProcessor(treeprocessors.Treeprocessor):  # type: ignore[misc]  # noqa: E501
    """
    Adds target="_blank" and rel="noopener" to external links.
    """

    def run(self, root: Element) -> None:
        for element in root.iter():
            if element.tag != "a":
                continue

            href = element.get("href", "")

            parsed_url = urlparse(href)

            if parsed_url.scheme in ["http", "https"]:
                element.set("target", "_blank")
                element.set("rel", "noopener")


class ExternalLinksExtension(Extension):  # type: ignore[misc]
    def extendMarkdown(self, md: Markdown) -> None:
        md.treeprocessors.register(
            ExternalLinksTreeProcessor(md),
            "external_links",
            -1000,
        )



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import inspect
import types
from typing import Any


def is_callable_method(obj: Any, attr: str) -> bool:
    if not hasattr(obj, attr):
        return False

    method = getattr(obj, attr)
    if inspect.isclass(obj) and not isinstance(method, (types.MethodType)):
        return False
    return callable(method)



# Copyright 2024 Marimo. All rights reserved.
import marimo

__generated_with = "0.3.8"
app = marimo.App()


@app.cell
def __(mo):
    mo.md(
        r"""
        # Pandas DataFrame: Extract values using regexp (regular expression)
        """
    )
    return


@app.cell
def __():
    import pandas as pd

    df = pd.DataFrame(
        {
            "request": ["GET /index.html?baz=3", "GET /foo.html?bar=1"],
        }
    )

    df["request"].str.extract(r"GET /([^?]+)\?", expand=True)
    return df, pd


@app.cell
def __():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import sys

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired as _NotRequired
else:
    from typing import NotRequired as _NotRequired

NotRequired = _NotRequired



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from marimo._messaging.mimetypes import KnownMimeType
from marimo._output.formatters.formatter_factory import FormatterFactory


class SeabornFormatter(FormatterFactory):
    @staticmethod
    def package_name() -> str:
        return "seaborn"

    def register(self) -> None:
        from typing import Any, cast

        # unused-ignore is needed since in development we may sometimes have
        # seaborn installed, in which case import-not-found is not applicable
        import seaborn  # type: ignore[import-not-found,import-untyped,unused-ignore] # noqa: E501

        from marimo._output import formatting
        from marimo._output.mime import MIME

        def _show_ax_or_subplots(grid: Any) -> tuple[KnownMimeType, str]:
            # Seaborn uses matplotlib under the hood: figures and axes are
            # instances of matplotlib's Artist class. (We've monkey pathed
            # `Artist` to implement the MIME protocol.)
            if hasattr(grid, "figure"):
                return cast(MIME, grid.figure)._mime_()
            if hasattr(grid, "ax"):
                return cast(MIME, grid.ax)._mime_()
            elif hasattr(grid, "axes"):
                return cast(MIME, grid.axes.flatten()[0])._mime_()
            else:
                return ("text/plain", str(grid))

        @formatting.formatter(seaborn.axisgrid.FacetGrid)
        def _show_facet_grid(
            fg: seaborn.axisgrid.FacetGrid,
        ) -> tuple[KnownMimeType, str]:
            return _show_ax_or_subplots(fg)

        @formatting.formatter(seaborn.axisgrid.PairGrid)
        def _show_pair_grid(
            pg: seaborn.axisgrid.PairGrid,
        ) -> tuple[KnownMimeType, str]:
            return _show_ax_or_subplots(pg)

        @formatting.formatter(seaborn.axisgrid.JointGrid)
        def _show_joint_grid(
            jg: seaborn.axisgrid.JointGrid,
        ) -> tuple[KnownMimeType, str]:
            return _show_ax_or_subplots(jg)



# Copyright 2024 Marimo. All rights reserved.
"""Message Types

Messages that the kernel sends to the frontend.
"""

from __future__ import annotations

import json
import sys
import time
from dataclasses import asdict, dataclass, field
from types import ModuleType
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from marimo import _loggers as loggers
from marimo._ast.app import _AppConfig
from marimo._ast.cell import CellConfig, CellId_t, RuntimeStateType
from marimo._data.models import ColumnSummary, DataTable, DataTableSource
from marimo._dependencies.dependencies import DependencyManager
from marimo._messaging.cell_output import CellChannel, CellOutput
from marimo._messaging.completion_option import CompletionOption
from marimo._messaging.errors import Error
from marimo._messaging.mimetypes import KnownMimeType
from marimo._messaging.streams import OUTPUT_MAX_BYTES
from marimo._messaging.types import Stream
from marimo._output.hypertext import Html
from marimo._plugins.core.json_encoder import WebComponentEncoder
from marimo._plugins.core.web_component import JSONType
from marimo._plugins.ui._core.ui_element import UIElement
from marimo._plugins.ui._impl.tables.utils import get_table_manager_or_none
from marimo._runtime.context import get_context
from marimo._runtime.layout.layout import LayoutConfig
from marimo._utils.platform import is_pyodide, is_windows

LOGGER = loggers.marimo_logger()


def serialize(datacls: Any) -> Dict[str, JSONType]:
    # TODO(akshayka): maybe serialize as bytes (JSON), not objects ...,
    # then `send_bytes` over connection ... to try to avoid pickling
    # issues
    try:
        # Try to serialize as a dataclass
        return cast(
            Dict[str, JSONType],
            asdict(datacls),
        )
    except Exception:
        # If that fails, try to serialize using the WebComponentEncoder
        return cast(
            Dict[str, JSONType],
            json.loads(WebComponentEncoder.json_dumps(datacls)),
        )


@dataclass
class Op:
    name: ClassVar[str]

    # TODO(akshayka): fix typing once mypy has stricter typing for asdict
    def broadcast(self, stream: Optional[Stream] = None) -> None:
        from marimo._runtime.context.types import ContextNotInitializedError

        if stream is None:
            try:
                ctx = get_context()
            except ContextNotInitializedError:
                LOGGER.debug("No context initialized.")
                return
            else:
                stream = ctx.stream

        try:
            stream.write(op=self.name, data=self.serialize())
        except Exception as e:
            LOGGER.exception(
                "Error serializing op %s: %s",
                self.__class__.__name__,
                e,
            )
            return

    def serialize(self) -> dict[str, Any]:
        return serialize(self)


@dataclass
class CellOp(Op):
    """Op to transition a cell.

    A CellOp's data has three optional fields:

    output       - a CellOutput
    console      - a CellOutput (console msg to append), or a list of
                   CellOutputs
    status       - execution status
    stale_inputs - whether the cell has stale inputs (variables, modules, ...)

    Omitting a field means that its value should be unchanged!

    And one required field:

    cell_id - the cell id
    """

    name: ClassVar[str] = "cell-op"
    cell_id: CellId_t
    output: Optional[CellOutput] = None
    console: Optional[Union[CellOutput, List[CellOutput]]] = None
    status: Optional[RuntimeStateType] = None
    stale_inputs: Optional[bool] = None
    timestamp: float = field(default_factory=lambda: time.time())

    @staticmethod
    def maybe_truncate_output(
        mimetype: KnownMimeType, data: str
    ) -> tuple[KnownMimeType, str]:
        if (size := sys.getsizeof(data)) > OUTPUT_MAX_BYTES:
            from marimo._output.md import md
            from marimo._plugins.stateless.callout import callout

            text = f"""
                <span class="text-error">**Your output is too large**</span>

                Your output is too large for marimo to show. It has a size
                of {size} bytes. Did you output this object by accident?

                If this limitation is a problem for you, you can configure
                the max output size with the environment variable
                `MARIMO_OUTPUT_MAX_BYTES`. For example, to increase
                the max output to 10 MB, use:

                ```
                export MARIMO_OUTPUT_MAX_BYTES=10_000_000
                ```

                Increasing the max output size may cause performance issues.
                If you run into problems, please reach out
                to us on [Discord](https://marimo.io/discord?ref=app) or
                [GitHub](https://github.com/marimo-team/marimo/issues).
                """

            warning = callout(
                md(text),
                kind="warn",
            )
            mimetype, data = warning._mime_()
        return mimetype, data

    @staticmethod
    def broadcast_output(
        channel: CellChannel,
        mimetype: KnownMimeType,
        data: str,
        cell_id: Optional[CellId_t],
        status: Optional[RuntimeStateType],
        stream: Stream | None = None,
    ) -> None:
        mimetype, data = CellOp.maybe_truncate_output(mimetype, data)
        cell_id = (
            cell_id if cell_id is not None else get_context().stream.cell_id
        )
        assert cell_id is not None
        CellOp(
            cell_id=cell_id,
            output=CellOutput(
                channel=channel,
                mimetype=mimetype,
                data=data,
            ),
            status=status,
        ).broadcast(stream=stream)

    @staticmethod
    def broadcast_empty_output(
        cell_id: Optional[CellId_t],
        status: Optional[RuntimeStateType],
        stream: Stream | None = None,
    ) -> None:
        cell_id = (
            cell_id if cell_id is not None else get_context().stream.cell_id
        )
        assert cell_id is not None
        CellOp(
            cell_id=cell_id,
            output=CellOutput(
                channel=CellChannel.OUTPUT,
                mimetype="text/plain",
                data="",
            ),
            status=status,
        ).broadcast(stream=stream)

    @staticmethod
    def broadcast_console_output(
        channel: CellChannel,
        mimetype: KnownMimeType,
        data: str,
        cell_id: Optional[CellId_t],
        status: Optional[RuntimeStateType],
        stream: Stream | None = None,
    ) -> None:
        mimetype, data = CellOp.maybe_truncate_output(mimetype, data)
        cell_id = (
            cell_id if cell_id is not None else get_context().stream.cell_id
        )
        assert cell_id is not None
        CellOp(
            cell_id=cell_id,
            console=CellOutput(
                channel=channel,
                mimetype=mimetype,
                data=data,
            ),
            status=status,
        ).broadcast(stream=stream)

    @staticmethod
    def broadcast_status(
        cell_id: CellId_t,
        status: RuntimeStateType,
        stream: Stream | None = None,
    ) -> None:
        if status != "running":
            CellOp(cell_id=cell_id, status=status).broadcast()
        else:
            # Console gets cleared on "running"
            CellOp(cell_id=cell_id, console=[], status=status).broadcast(
                stream=stream
            )

    @staticmethod
    def broadcast_error(
        data: Sequence[Error],
        clear_console: bool,
        cell_id: CellId_t,
    ) -> None:
        console: Optional[list[CellOutput]] = [] if clear_console else None
        CellOp(
            cell_id=cell_id,
            output=CellOutput(
                channel=CellChannel.MARIMO_ERROR,
                mimetype="application/vnd.marimo+error",
                data=data,
            ),
            console=console,
            status=None,
        ).broadcast()

    @staticmethod
    def broadcast_stale(
        cell_id: CellId_t, stale: bool, stream: Stream | None = None
    ) -> None:
        CellOp(cell_id=cell_id, stale_inputs=stale).broadcast(stream)


@dataclass
class HumanReadableStatus:
    """Human-readable status."""

    code: Literal["ok", "error"]
    title: Union[str, None] = None
    message: Union[str, None] = None


@dataclass
class FunctionCallResult(Op):
    """Result of calling a function."""

    name: ClassVar[str] = "function-call-result"

    function_call_id: str
    return_value: JSONType
    status: HumanReadableStatus

    def __post_init__(self) -> None:
        # We want to serialize the return_value using our custom JSON encoder
        try:
            self.return_value = json.loads(
                WebComponentEncoder.json_dumps(self.return_value)
            )
        except Exception as e:
            LOGGER.exception(
                "Error serializing function call result %s: %s",
                self.__class__.__name__,
                e,
            )

    def serialize(self) -> dict[str, Any]:
        try:
            return serialize(self)
        except Exception as e:
            LOGGER.exception(
                "Error serializing function call result %s: %s",
                self.__class__.__name__,
                e,
            )
            return serialize(
                FunctionCallResult(
                    function_call_id=self.function_call_id,
                    return_value=None,
                    status=HumanReadableStatus(
                        code="error",
                        title="Error calling function",
                        message="Failed to serialize function call result",
                    ),
                )
            )


@dataclass
class RemoveUIElements(Op):
    """Invalidate UI elements for a given cell."""

    name: ClassVar[str] = "remove-ui-elements"
    cell_id: CellId_t


@dataclass
class SendUIElementMessage(Op):
    """Send a message to a UI element."""

    name: ClassVar[str] = "send-ui-element-message"
    ui_element: str
    message: Dict[str, object]
    buffers: Optional[Sequence[str]]


@dataclass
class Interrupted(Op):
    """Written when the kernel is interrupted by the user."""

    name: ClassVar[str] = "interrupted"


@dataclass
class CompletedRun(Op):
    """Written on run completion (of submitted cells and their descendants."""

    name: ClassVar[str] = "completed-run"


@dataclass
class KernelCapabilities:
    sql: bool = False
    terminal: bool = False

    def __post_init__(self) -> None:
        self.sql = DependencyManager.duckdb.has_at_version(min_version="1.0.0")
        # Only available in mac/linux
        self.terminal = not is_windows() and not is_pyodide()


@dataclass
class KernelReady(Op):
    """Kernel is ready for execution."""

    name: ClassVar[str] = "kernel-ready"
    cell_ids: Tuple[CellId_t, ...]
    codes: Tuple[str, ...]
    names: Tuple[str, ...]
    layout: Optional[LayoutConfig]
    configs: Tuple[CellConfig, ...]
    # Whether the kernel was resumed from a previous session
    resumed: bool
    # If the kernel was resumed, the values of the UI elements
    ui_values: Optional[Dict[str, JSONType]]
    # If the kernel was resumed, the last executed code for each cell
    last_executed_code: Optional[Dict[CellId_t, str]]
    # If the kernel was resumed, the last execution time for each cell
    last_execution_time: Optional[Dict[CellId_t, float]]
    # App config
    app_config: _AppConfig
    # Whether the kernel is kiosk mode
    kiosk: bool
    # Kernel capabilities
    capabilities: KernelCapabilities


@dataclass
class CompletionResult(Op):
    """Code completion result."""

    name: ClassVar[str] = "completion-result"
    completion_id: str
    prefix_length: int
    options: List[CompletionOption]


@dataclass
class Alert(Op):
    name: ClassVar[str] = "alert"
    title: str
    # description may be HTML
    description: str
    variant: Optional[Literal["danger"]] = None


@dataclass
class MissingPackageAlert(Op):
    name: ClassVar[str] = "missing-package-alert"
    packages: List[str]
    isolated: bool


# package name => installation status
PackageStatusType = Dict[
    str, Literal["queued", "installing", "installed", "failed"]
]


@dataclass
class InstallingPackageAlert(Op):
    name: ClassVar[str] = "installing-package-alert"
    packages: PackageStatusType


@dataclass
class Reconnected(Op):
    name: ClassVar[str] = "reconnected"


@dataclass
class Banner(Op):
    name: ClassVar[str] = "banner"
    title: str
    # description may be HTML
    description: str
    variant: Optional[Literal["danger"]] = None
    action: Optional[Literal["restart"]] = None


@dataclass
class Reload(Op):
    name: ClassVar[str] = "reload"


@dataclass
class VariableDeclaration:
    name: str
    declared_by: List[CellId_t]
    used_by: List[CellId_t]


@dataclass
class VariableValue:
    name: str
    value: Optional[str]
    datatype: Optional[str]

    def __init__(
        self, name: str, value: object, datatype: Optional[str] = None
    ) -> None:
        self.name = name

        # Defensively try-catch attribute accesses, which could raise
        # exceptions
        # If datatype is already defined, don't try to infer it
        if datatype is None:
            try:
                self.datatype = (
                    type(value).__name__ if value is not None else None
                )
            except Exception:
                self.datatype = datatype
        else:
            self.datatype = datatype

        try:
            self.value = self._format_value(value)
        except Exception:
            self.value = None

    def _stringify(self, value: object) -> str:
        try:
            # HACK: We pretty-print tables to avoid str(ibis_table)
            # which can be very slow when `ibis.options.interactive = True`
            table_manager = get_table_manager_or_none(value)
            if table_manager is not None:
                return str(table_manager)
            else:
                return str(value)[:50]

            return str(value)[:50]
        except BaseException:
            # Catch-all: some libraries like Polars have bugs and raise
            # BaseExceptions, which shouldn't crash the kernel
            return "<UNKNOWN>"

    def _format_value(self, value: object) -> str:
        resolved = value
        if isinstance(value, UIElement):
            resolved = value.value
        elif isinstance(value, Html):
            resolved = value.text
        elif isinstance(value, ModuleType):
            resolved = value.__name__
        return self._stringify(resolved)


@dataclass
class Variables(Op):
    """List of variable declarations."""

    name: ClassVar[str] = "variables"
    variables: List[VariableDeclaration]


@dataclass
class VariableValues(Op):
    """List of variables and their types/values."""

    name: ClassVar[str] = "variable-values"
    variables: List[VariableValue]


@dataclass
class Datasets(Op):
    """List of datasets."""

    name: ClassVar[str] = "datasets"
    tables: List[DataTable]
    clear_channel: Optional[DataTableSource] = None


@dataclass
class DataColumnPreview(Op):
    """Preview of a column in a dataset."""

    name: ClassVar[str] = "data-column-preview"
    table_name: str
    column_name: str
    chart_spec: Optional[str] = None
    chart_max_rows_errors: bool = False
    chart_code: Optional[str] = None
    error: Optional[str] = None
    summary: Optional[ColumnSummary] = None


@dataclass
class QueryParamsSet(Op):
    """Set query parameters."""

    name: ClassVar[str] = "query-params-set"
    key: str
    value: Union[str, List[str]]


@dataclass
class QueryParamsAppend(Op):
    name: ClassVar[str] = "query-params-append"
    key: str
    value: str


@dataclass
class QueryParamsDelete(Op):
    name: ClassVar[str] = "query-params-delete"
    key: str
    # If value is None, delete all values for the key
    # If a value is provided, only that value is deleted
    value: Optional[str]


@dataclass
class QueryParamsClear(Op):
    # Clear all query parameters
    name: ClassVar[str] = "query-params-clear"


@dataclass
class FocusCell(Op):
    name: ClassVar[str] = "focus-cell"
    cell_id: CellId_t


@dataclass
class UpdateCellCodes(Op):
    name: ClassVar[str] = "update-cell-codes"
    cell_ids: List[CellId_t]
    codes: List[str]


@dataclass
class UpdateCellIdsRequest(Op):
    """
    Update the cell ID ordering of the cells in the notebook.

    Right now we send the entire list of cell IDs,
    but in the future we might want to send change-deltas.
    """

    name: ClassVar[str] = "update-cell-ids"
    cell_ids: List[CellId_t]


MessageOperation = Union[
    # Cell operations
    CellOp,
    FunctionCallResult,
    SendUIElementMessage,
    RemoveUIElements,
    # Notebook operations
    Reload,
    Reconnected,
    Interrupted,
    CompletedRun,
    KernelReady,
    # Editor operations
    CompletionResult,
    # Alerts
    Alert,
    Banner,
    MissingPackageAlert,
    InstallingPackageAlert,
    # Variables
    Variables,
    VariableValues,
    # Query params
    QueryParamsSet,
    QueryParamsAppend,
    QueryParamsDelete,
    QueryParamsClear,
    # Datasets
    Datasets,
    DataColumnPreview,
    # Kiosk specific
    FocusCell,
    UpdateCellCodes,
    UpdateCellIdsRequest,
]



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from typing import Any, Generic, List, TypeVar

from marimo._dependencies.dependencies import DependencyManager
from marimo._plugins.ui._impl.dataframes.transforms.handlers import (
    IbisTransformHandler,
    PandasTransformHandler,
    PolarsTransformHandler,
)
from marimo._plugins.ui._impl.dataframes.transforms.types import (
    Transform,
    Transformations,
    TransformHandler,
    TransformType,
)
from marimo._utils.assert_never import assert_never

T = TypeVar("T")


def _handle(df: T, handler: TransformHandler[T], transform: Transform) -> T:
    if transform.type is TransformType.COLUMN_CONVERSION:
        return handler.handle_column_conversion(df, transform)
    elif transform.type is TransformType.RENAME_COLUMN:
        return handler.handle_rename_column(df, transform)
    elif transform.type is TransformType.SORT_COLUMN:
        return handler.handle_sort_column(df, transform)
    elif transform.type is TransformType.FILTER_ROWS:
        return handler.handle_filter_rows(df, transform)
    elif transform.type is TransformType.GROUP_BY:
        return handler.handle_group_by(df, transform)
    elif transform.type is TransformType.AGGREGATE:
        return handler.handle_aggregate(df, transform)
    elif transform.type is TransformType.SELECT_COLUMNS:
        return handler.handle_select_columns(df, transform)
    elif transform.type is TransformType.SHUFFLE_ROWS:
        return handler.handle_shuffle_rows(df, transform)
    elif transform.type is TransformType.SAMPLE_ROWS:
        return handler.handle_sample_rows(df, transform)
    elif transform.type is TransformType.EXPLODE_COLUMNS:
        return handler.handle_explode_columns(df, transform)
    elif transform.type is TransformType.EXPAND_DICT:
        return handler.handle_expand_dict(df, transform)
    else:
        assert_never(transform.type)


def _apply_transforms(
    df: T, handler: TransformHandler[T], transforms: Transformations
) -> T:
    if not transforms.transforms:
        return df
    for transform in transforms.transforms:
        df = _handle(df, handler, transform)
    return df


def get_handler_for_dataframe(
    df: Any,
) -> TransformHandler[Any]:
    """
    Gets the handler for the given dataframe.

    raises ValueError if the dataframe type is not supported.
    """
    if DependencyManager.pandas.imported():
        import pandas as pd

        if isinstance(df, pd.DataFrame):
            return PandasTransformHandler()
    if DependencyManager.polars.imported():
        import polars as pl

        if isinstance(df, pl.DataFrame):
            return PolarsTransformHandler()

    if DependencyManager.ibis.imported():
        import ibis  # type: ignore

        if isinstance(df, ibis.Table):
            return IbisTransformHandler()

    if DependencyManager.narwhals.imported():
        import narwhals as nw

        if isinstance(df, nw.DataFrame):
            return get_handler_for_dataframe(df.to_native())

    raise ValueError(
        "Unsupported dataframe type. Must be Pandas or Polars."
        f" Got: {type(df)}"
    )


class TransformsContainer(Generic[T]):
    """
    Keeps internal state of the last transformation applied to the dataframe.
    So that we can incrementally apply transformations.
    """

    def __init__(self, df: T, handler: TransformHandler[T]) -> None:
        self._original_df = df
        # The dataframe for the given transform.
        self._snapshot_df = df
        self._handler = handler
        self._transforms: List[Transform] = []

    def apply(self, transform: Transformations) -> T:
        """
        Applies the given transformations to the dataframe.
        """
        # If the new transformations are a superset of the existing ones,
        # then we can just apply the new ones to the snapshot dataframe.
        if self._is_superset(transform):
            transforms_to_apply = self._get_next_transformations(transform)
            self._snapshot_df = _apply_transforms(
                self._snapshot_df, self._handler, transforms_to_apply
            )
            self._transforms = transform.transforms
            return self._snapshot_df

        # If the new transformations are not a superset of the existing ones,
        # then we need to start from the original dataframe.
        else:
            self._snapshot_df = _apply_transforms(
                self._original_df, self._handler, transform
            )
            self._transforms = transform.transforms
            return self._snapshot_df

    def _is_superset(self, transforms: Transformations) -> bool:
        """
        Checks if the new transformations are a superset of the existing ones.
        """
        if not self._transforms:
            return False

        # If the new transformations are smaller than the existing ones,
        # then it's not a superset.
        if len(self._transforms) > len(transforms.transforms):
            return False

        for i, transform in enumerate(self._transforms):
            if transform != transforms.transforms[i]:
                return False

        return True

    def _get_next_transformations(
        self, transforms: Transformations
    ) -> Transformations:
        """
        Gets the next transformations to apply.
        """
        if self._is_superset(transforms):
            return Transformations(
                transforms.transforms[len(self._transforms) :]
            )
        else:
            return transforms



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import abc
import mimetypes
from dataclasses import dataclass
from typing import List, Literal, Optional, TypedDict


class ChatAttachmentDict(TypedDict):
    url: str
    content_type: Optional[str]
    name: Optional[str]


class ChatMessageDict(TypedDict):
    role: Literal["user", "assistant", "system"]
    content: str
    attachments: Optional[List[ChatAttachmentDict]]


class ChatModelConfigDict(TypedDict, total=False):
    max_tokens: Optional[int]
    temperature: Optional[float]
    top_p: Optional[float]
    top_k: Optional[int]
    frequency_penalty: Optional[float]
    presence_penalty: Optional[float]


# NOTE: The following classes are public API.
# Any changes must be backwards compatible.


@dataclass
class ChatAttachment:
    # The URL of the attachment. It can either be a URL to a hosted file or a
    # [Data URL](https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/Data_URLs).
    url: str

    # The name of the attachment, usually the file name.
    name: str = "attachment"

    # A string indicating the [media type](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type).
    # By default, it's extracted from the pathname's extension.
    content_type: Optional[str] = None

    def __post_init__(self) -> None:
        if self.content_type is None:
            self.content_type = mimetypes.guess_type(self.url)[0]


@dataclass
class ChatMessage:
    """
    A message in a chat.
    """

    # The role of the message.
    role: Literal["user", "assistant", "system"]

    # The content of the message.
    content: object

    # Optional attachments to the message.
    attachments: Optional[List[ChatAttachment]] = None


@dataclass
class ChatModelConfig:
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None


class ChatModel(abc.ABC):
    @abc.abstractmethod
    def __call__(
        self, messages: List[ChatMessage], config: ChatModelConfig
    ) -> object:
        pass



# Copyright 2024 Marimo. All rights reserved.
from marimo._output.formatting import as_html
from marimo._output.hypertext import Html
from marimo._output.rich_help import mddoc


@mddoc
def center(item: object) -> Html:
    """Center an item.

    **Returns.**

    A centered `Html` object.
    """
    return as_html(item).center()


@mddoc
def left(item: object) -> Html:
    """Left-justify an item.

    **Returns.**

    A left-justified `Html` object.
    """
    return as_html(item).left()


@mddoc
def right(item: object) -> Html:
    """Right-justify an item.

    **Returns.**

    A right-justified `Html` object.
    """
    return as_html(item).right()



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from typing import TYPE_CHECKING

from marimo._messaging.mimetypes import KnownMimeType
from marimo._output.formatters.formatter_factory import FormatterFactory

if TYPE_CHECKING:
    import matplotlib.pyplot as plt  # type: ignore
    import numpy as np  # type: ignore
    from matplotlib.figure import Figure  # type: ignore


class ArviZFormatter(FormatterFactory):
    @staticmethod
    def package_name() -> str:
        return "arviz"

    def register(self) -> None:
        import arviz as az  # type: ignore
        import matplotlib.pyplot as plt  # type: ignore
        import numpy as np  # type: ignore

        from marimo._output import formatting

        @formatting.formatter(az.InferenceData)  # type: ignore
        def _format_inference_data(
            data: az.InferenceData,  # type: ignore
        ) -> tuple[KnownMimeType, str]:
            return ("text/plain", str(data))

        @formatting.formatter(np.ndarray)  # type: ignore
        def _format_ndarray(
            arr: np.ndarray,  # type: ignore
        ) -> tuple[KnownMimeType, str]:
            return self.format_numpy_axes(arr)

        @formatting.formatter(plt.Figure)  # type: ignore
        def _format_figure(
            fig: plt.Figure,  # type: ignore
        ) -> tuple[KnownMimeType, str]:
            return self.format_figure(fig)

    @classmethod
    def format_numpy_axes(cls, arr: np.ndarray) -> tuple[KnownMimeType, str]:  # type: ignore
        import matplotlib.pyplot as plt  # type: ignore

        # Check if array contains axes (to render plots) or not
        if arr.dtype == object and cls._contains_axes(arr):
            fig = plt.gcf()
            if fig.get_axes():  # Only process if there are axes to show
                axes_info = cls._get_axes_info(fig)
                plot_html = cls._get_plot_html(fig)
                plt.close(fig)  # Safely close the figure after saving
                combined_html = f"<pre>{axes_info}</pre><br>{plot_html}"
                return ("text/html", combined_html)
        # Fallback to plain text if no axes or plot are present
        return ("text/plain", str(arr))

    @staticmethod
    def _contains_axes(arr: np.ndarray) -> bool:  # type: ignore
        from matplotlib.axes import Axes  # type: ignore

        """
        Check if the numpy array contains any matplotlib Axes objects.
        To ensure performance for large arrays, we limit the check to the
        first 100 items. This should be sufficient for most use cases
        while avoiding excessive computation time.
        """
        # Cap the number of items to check for performance reasons
        MAX_ITEMS_TO_CHECK = 100

        if arr.ndim == 1:
            # For 1D arrays, check up to MAX_ITEMS_TO_CHECK items
            return any(
                isinstance(item, Axes) for item in arr[:MAX_ITEMS_TO_CHECK]
            )
        elif arr.ndim == 2:
            # For 2D arrays, check up to MAX_ITEMS_TO_CHECK items in total
            items_checked = 0
            for row in arr:
                for item in row:
                    if isinstance(item, Axes):
                        return True
                    items_checked += 1
                    if items_checked >= MAX_ITEMS_TO_CHECK:
                        return False
        return False

    @staticmethod
    def _get_axes_info(fig: Figure) -> str:  # type: ignore
        axes_info = []
        for _, ax in enumerate(fig.axes):
            bbox = ax.get_position()
            axes_info.append(
                f"Axes({bbox.x0:.3f},{bbox.y0:.3f};"
                f"{bbox.width:.3f}x{bbox.height:.3f})"
            )
        return "\n".join(axes_info)

    @staticmethod
    def _get_plot_html(fig: Figure) -> str:  # type: ignore
        import base64
        from io import BytesIO

        buf = BytesIO()
        fig.savefig(buf, format="png", bbox_inches="tight")  # Retain default
        data = base64.b64encode(buf.getbuffer()).decode("ascii")
        return f"<img src='data:image/png;base64,{data}'/>"

    @classmethod
    def format_figure(cls, fig: Figure) -> tuple[KnownMimeType, str]:  # type: ignore
        import matplotlib.pyplot as plt  # type: ignore

        axes_info = cls._get_axes_info(fig)
        plot_html = cls._get_plot_html(fig)
        plt.close(fig)
        combined_html = f"<pre>{axes_info}</pre><br>{plot_html}"
        return ("text/html", combined_html)



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations


# Mapping from a module name to the corresponding package name
# on PyPI
def module_name_to_pypi_name() -> dict[str, str]:
    return {
        "cv2": "opencv-python",
        "shopify": "ShopifyAPI",
        "bigquery": "google-cloud-bigquery",
        "carbon3d": "carbon3d-client",
        "telegram": "python-telegram-bot",
        "AFQ": "pyAFQ",
        "AG_fft_tools": "agpy",
        "screen": "pexpect",
        "Adafruit": "Adafruit-Libraries",
        "webdav": "Zope2",
        "Asterisk": "py-Asterisk",
        "BB_jekyll_hook": "bitbucket-jekyll-hook",
        "Banzai": "Banzai-NGS",
        "BeautifulSoupTests": "BeautifulSoup",
        "BioSQL": "biopython",
        "BuildbotStatusShields": "BuildbotEightStatusShields",
        "MethodObject": "ExtensionClass",
        "Crypto": "pycryptodome",
        "Cryptodome": "pycryptodomex",
        "FiftyOneDegrees": "51degrees-mobile-detector-v3-wrapper",
        "functional": "pyfunctional",
        "GeoBases": "GeoBasesDev",
        "IPython": "ipython",
        "Kittens": "astro-kittens",
        "Levenshtein": "python-Levenshtein",
        "MySQLdb": "MySQL-python",
        "OpenGL": "PyOpenGL",
        "OpenSSL": "pyOpenSSL",
        "PIL": "Pillow",
        "PyWCSTools": "astLib",
        "Pyxides": "astro-pyxis",
        "pysideuic": "PySide",
        "S3": "s3cmd",
        "SCons": "pystick",
        "Stemmer": "PyStemmer",
        "TopZooTools": "topzootools",
        "TreeDisplay": "DocumentTemplate",
        "WorkingWithDocumentConversion": "aspose-pdf-java-for-python",
        "aadb": "auto-adjust-display-brightness",
        "abakaffe": "abakaffe-cli",
        "abiosgaming": "abiosgaming.py",
        "abiquo": "abiquo-api",
        "abl": "abl.vpath",
        "abo": "abo-generator",
        "abris_transform": "abris",
        "abstract": "abstract.jwrotator",
        "abu": "abu.admin",
        "ac_flask": "AC-Flask-HipChat",
        "acg": "anikom15",
        "acme": "acme.hello",
        "acted": "acted.projects",
        "action": "ActionServer",
        "actionbar": "actionbar.panel",
        "activehomed": "afn",
        "activepapers": "ActivePapers.Py",
        "address_book": "address-book-lansry",
        "adi": "adi.trash",
        "adict": "aDict2",
        "aditam": "aditam.core",
        "adiumsh": "adium-sh",
        "adjector": "AdjectorTracPlugin",
        "adkit": "Banner-Ad-Toolkit",
        "admin_tools": "django-admin-tools",
        "adminishcategories": "adminish-categories",
        "adminsortable": "django-admin-sortable",
        "adspygoogle": "adspygoogle.adwords",
        "advancedcaching": "agtl",
        "adytum": "Adytum-PyMonitor",
        "affinitic": "affinitic.zamqp",
        "afpy": "afpy.xap",
        "agatesql": "agate-sql",
        "ageliaco": "ageliaco.recipe.csvconfig",
        "agent_http": "agent.http",
        "agora": "Agora-Service-Provider",
        "agoraplex": "agoraplex.themes.sphinx",
        "agsci": "agsci.blognewsletter",
        "agx": "agx.transform.xmi2uml",
        "aimes": "aimes.skeleton",
        "aio": "aio.signals",
        "aiohs2": "aio-hs2",
        "aioroutes": "aio-routes",
        "aios3": "aio-s3",
        "airbrake": "airbrake-flask",
        "airship": "airship-steamcloud",
        "akamai": "edgegrid-python",
        "alation": "alation-api",
        "alba_client": "alba-client-python",
        "alburnum": "alburnum-maas-client",
        "alchemist": "alchemist.ui",
        "alchemyapi": "alchemyapi-python",
        "alerta": "alerta-server",
        "alexandria_upload": "Alexandria-Upload-Utils",
        "alibaba": "alibaba-python-sdk",
        "aliyun": "aliyun-python-sdk",
        "aliyuncli": "alicloudcli",
        "aliyunsdkacs": "aliyun-python-sdk-acs",
        "aliyunsdkbatchcompute": "aliyun-python-sdk-batchcompute",
        "aliyunsdkbsn": "aliyun-python-sdk-bsn",
        "aliyunsdkbss": "aliyun-python-sdk-bss",
        "aliyunsdkcdn": "aliyun-python-sdk-cdn",
        "aliyunsdkcms": "aliyun-python-sdk-cms",
        "aliyunsdkcore": "aliyun-python-sdk-core",
        "aliyunsdkcrm": "aliyun-python-sdk-crm",
        "aliyunsdkcs": "aliyun-python-sdk-cs",
        "aliyunsdkdrds": "aliyun-python-sdk-drds",
        "aliyunsdkecs": "aliyun-python-sdk-ecs",
        "aliyunsdkess": "aliyun-python-sdk-ess",
        "aliyunsdkft": "aliyun-python-sdk-ft",
        "aliyunsdkmts": "aliyun-python-sdk-mts",
        "aliyunsdkocs": "aliyun-python-sdk-ocs",
        "aliyunsdkoms": "aliyun-python-sdk-oms",
        "aliyunsdkossadmin": "aliyun-python-sdk-ossadmin",
        "aliyunsdkr-kvstore": "aliyun-python-sdk-r-kvstore",
        "aliyunsdkram": "aliyun-python-sdk-ram",
        "aliyunsdkrds": "aliyun-python-sdk-rds",
        "aliyunsdkrisk": "aliyun-python-sdk-risk",
        "aliyunsdkros": "aliyun-python-sdk-ros",
        "aliyunsdkslb": "aliyun-python-sdk-slb",
        "aliyunsdksts": "aliyun-python-sdk-sts",
        "aliyunsdkubsms": "aliyun-python-sdk-ubsms",
        "aliyunsdkyundun": "aliyun-python-sdk-yundun",
        "allattachments": "AllAttachmentsMacro",
        "allocine": "allocine-wrapper",
        "allowedsites": "django-allowedsites",
        "alm": "alm.solrindex",
        "aloft": "aloft.py",
        "alpacalib": "alpaca",
        "alphabetic": "alphabetic-simple",
        "alphasms": "alphasms-client",
        "altered": "altered.states",
        "alterootheme": "alterootheme.lazydays",
        "alurinium": "alurinium-image-processing",
        "alxlib": "alx",
        "amara3": "amara3-xml",
        "amazon": "python-amazon-simple-product-api",
        "ambikesh1349-1": "ambikesh1349-1",
        "ambilight": "AmbilightParty",
        "amifs": "amifs-core",
        "amiorganizer": "ami-organizer",
        "amitu": "amitu-zutils",
        "amltlearn": "AMLT-learn",
        "amocrm": "amocrm-api",
        "amqpdispatcher": "amqp-dispatcher",
        "amqpstorm": "AMQP-Storm",
        "analytics": "analytics-python",
        "analyzedir": "AnalyzeDirectory",
        "ancientsolutions": "ancientsolutions-crypttools",
        "anderson_paginator": "anderson.paginator",
        "android_clean_app": "android-resource-remover",
        "anel_power_control": "AnelPowerControl",
        "angus": "angus-sdk-python",
        "annalist_root": "Annalist",
        "annogesiclib": "ANNOgesic",
        "ansible-role-apply": "ansible-role-apply",
        "ansibledebugger": "ansible-playbook-debugger",
        "ansibledocgen": "ansible-docgen",
        "ansibleflow": "ansible-flow",
        "ansibleinventorygrapher": "ansible-inventory-grapher",
        "ansiblelint": "ansible-lint",
        "ansiblerolesgraph": "ansible-roles-graph",
        "ansibletools": "ansible-tools",
        "anthill": "anthill.tal.macrorenderer",
        "anthrax": "AnthraxImage",
        "antisphinx": "antiweb",
        "antispoofing": "antispoofing.evaluation",
        "antlr4": "antlr4-python-alt",
        "anybox": "anybox.scripts.odoo",
        "googleapiclient": "google-api-python-client",
        "apitools": "google-apitools",
        "apm": "arpm",
        "app_data": "django-appdata",
        "appconf": "django-appconf",
        "appd": "AppDynamicsREST",
        "appdynamics_bindeps": "appdynamics-bindeps-osx-x64",
        "appdynamics_proxysupport": "appdynamics-proxysupport-osx-x64",
        "appium": "Appium-Python-Client",
        "appliapps": "applibase",
        "appserver": "broadwick",
        "archetypes": "archetypes.schemaextender",
        "arm": "ansible-role-manager",
        "armor": "armor-api",
        "armstrong": "armstrong.utils.celery",
        "arstecnica": "arstecnica.sqlalchemy.async",
        "article-downloader": "article-downloader",
        "artifactcli": "artifact-cli",
        "arvados": "arvados-python-client",
        "arvados_cwl": "arvados-cwl-runner",
        "arvnodeman": "arvados-node-manager",
        "asana_to_github": "AsanaToGithub",
        "asciibinary": "AsciiBinaryConverter",
        "asd": "AdvancedSearchDiscovery",
        "askbot": "askbot-tuanpa",
        "asnhistory": "asnhistory-redis",
        "aspen_jinja2_renderer": "aspen-jinja2",
        "aspen_tornado_engine": "aspen-tornado",
        "asprise_ocr_api": "asprise-ocr-sdk-python-api",
        "aspy": "aspy.yaml",
        "asterisk": "asterisk-ami",
        "asts": "add-asts",
        "asymmetricbase": "asymmetricbase.utils",
        "asyncirc": "asyncio-irc",
        "asyncmongoorm": "asyncmongoorm-je",
        "asyncssh": "asyncssh-unofficial",
        "athletelist": "athletelistyy",
        "atm": "automium",
        "atmosphere": "atmosphere-python-client",
        "atom": "gdata",
        "atomic": "AtomicWrite",
        "atomisator": "atomisator.readers",
        "atreal": "atreal.usersinout",
        "atsim": "atsim.potentials",
        "attractsdk": "attract-sdk",
        "audio": "audio.wave",
        "aufrefer": "auf-refer",
        "auslfe": "auslfe.formonline.content",
        "auspost": "auspost-apis",
        "auth0": "auth0-python",
        "auth_server_client": "AuthServerClient",
        "authorize": "AuthorizeSauce",
        "authzpolicy": "AuthzPolicyPlugin",
        "autobahn": "autobahn-rce",
        "avatar": "geonode-avatar",
        "awebview": "android-webview",
        "azure": "azure-storage",
        "b2gcommands": "b2g-commands",
        "b2gperf": "b2gperf-v2.2",
        "b2gpopulate": "b2gpopulate-v2.2",
        "b3j0f": "b3j0f.utils",
        "babel": "Babel",
        "babelglade": "BabelGladeExtractor",
        "backplane": "backplane2-pyclient",
        "backport_abcoll": "backport-collections",
        "backports": "backports.statistics",
        "badgekit": "badgekit-api-client",
        "badlinks": "BadLinksPlugin",
        "bael": "bael.project",
        "baidu": "baidupy",
        "balrog": "buildtools",
        "baluhn": "baluhn-redux",
        "bamboo": "bamboo-server",
        "bambu": "bambu-sites",
        "banana": "banana.maya",
        "bang": "bangtext",
        "barcode": "barcode-generator",
        "bark": "bark-ssg",
        "barking_owl": "BarkingOwl",
        "bart": "bart-py",
        "basalt": "basalt-tasks",
        "base62": "base-62",
        "basemap": "basemap-Jim",
        "bash": "bash-toolbelt",
        "bashutils": "Python-Bash-Utils",
        "basic_http": "BasicHttp",
        "basil": "basil-daq",
        "batchapps": "azure-batch-apps",
        "bcrypt": "python-bcrypt",
        "beaker": "Beaker",
        "beetsplug": "beets",
        "begin": "begins",
        "benchit": "bench-it",
        "beproud": "beproud.utils",
        "bfillings": "burrito-fillings",
        "pilot": "BigJob",
        "billboard": "billboard.py",
        "binstar_build_client": "anaconda-build",
        "binstar_client": "anaconda-client",
        "biocommons": "biocommons.dev",
        "birdhousebuilder": "birdhousebuilder.recipe.supervisor",
        "blender26-meshio": "pymeshio",
        "borg": "borg.localrole",
        "bow": "bagofwords",
        "bpdb": "bpython",
        "bqapi": "bisque-api",
        "braces": "django-braces",
        "briefscaster": "briefs-caster",
        "brisa_media_server/plugins": "brisa-media-server-plugins",
        "brkt_requests": "brkt-sdk",
        "broadcastlogging": "broadcast-logging",
        "brocadetool": "brocade-tool",
        "bronto": "bronto-python",
        "brownie": "Brownie",
        "browsermobproxy": "browsermob-proxy",
        "brubeckmysql": "brubeck-mysql",
        "brubeckoauth": "brubeck-oauth",
        "brubeckservice": "brubeck-service",
        "brubeckuploader": "brubeck-uploader",
        "bs4": "beautifulsoup4",
        "gridfs": "pymongo",
        "bst": "bst.pygasus.wsgi",
        "btable": "btable-py",
        "btapi": "bananatag-api",
        "btceapi": "btce-api",
        "btcebot": "btce-bot",
        "btsync": "btsync.py",
        "buck": "buck.pprint",
        "bud": "bud.nospam",
        "budy": "budy-api",
        "buffer": "buffer-alpaca",
        "buggd": "bug.gd",
        "bugle": "bugle-sites",
        "bugspots": "bug-spots",
        "bugzilla": "python-bugzilla",
        "bugzscout": "bugzscout-py",
        "buildTools": "ajk-ios-buildTools",
        "buildnotifylib": "BuildNotify",
        "buildout": "buildout.variables",
        "buildslave": "buildbot-slave",
        "xmlrpc": "pies2overrides",
        "bumper": "bumper-lib",
        "bumple": "bumple-downloader",
        "bundesliga": "bundesliga-cli",
        "bundlemaker": "bundlemanager",
        "burpui": "burp-ui",
        "busyflow": "busyflow.pivotal",
        "buttercms-django": "buttercms-django",
        "buzz": "buzz-python-client",
        "bvc": "buildout-versions-checker",
        "bvggrabber": "bvg-grabber",
        "byond": "BYONDTools",
        "bzETL": "Bugzilla-ETL",
        "bzlib": "bugzillatools",
        "bzrlib": "bzr-pqm",
        "c2c": "c2c.versions",
        "c2c_recipe_facts": "c2c.recipe.facts",
        "cabalgata": "cabalgata-zookeeper",
        "cache_utils": "django-cache-utils",
        "captcha": "django-recaptcha",
        "cartridge": "Cartridge",
        "cassandra": "cassandra-driver",
        "cassandralauncher": "CassandraLauncher",
        "cc42": "42qucc",
        "cerberus": "Cerberus",
        "cfnlint": "cfn-lint",
        "chameleon": "Chameleon",
        "charmtools": "charm-tools",
        "chef": "PyChef",
        "chip8": "c8d",
        "cjson": "python-cjson",
        "classytags": "django-classy-tags",
        "cloghandler": "ConcurrentLogHandler",
        "clonevirtualenv": "virtualenv-clone",
        "cloud-insight": "al-cloudinsight",
        "cloud_admin": "adminapi",
        "cloudservers": "python-cloudservers",
        "tasksitter": "cerebrod",
        "cms": "django-cms",
        "colander": "ba-colander",
        "colors": "ansicolors",
        "compile": "bf-lc3",
        "compose": "docker-compose",
        "compressor": "django-compressor",
        "concurrent": "futures",
        "configargparse": "ConfigArgParse",
        "contracts": "PyContracts",
        "weblogolib": "weblogo",
        "couchapp": "Couchapp",
        "couchdb": "CouchDB",
        "couchdbcurl": "couchdb-python-curl",
        "courseradownloader": "coursera-dl",
        "cow": "cow-framework",
        "creole": "python-creole",
        "creoleparser": "Creoleparser",
        "crispy_forms": "django-crispy-forms",
        "crontab": "python-crontab",
        "ctff": "tff",
        "cups": "pycups",
        "curator": "elasticsearch-curator",
        "curl": "pycurl",
        "daemon": "python-daemon",
        "dagger": "dagger-io",
        "dare": "DARE",
        "dateutil": "python-dateutil",
        "dawg": "DAWG",
        "debian": "python-debian",
        "decouple": "python-decouple",
        "demo": "webunit",
        "pysynth_samp": "PySynth",
        "deployer": "juju-deployer",
        "depot": "filedepot",
        "dgis": "2gis",
        "dhtmlparser": "pyDHTMLParser",
        "digitalocean": "python-digitalocean",
        "discord": "discord.py",
        "distribute_setup": "ez-setup",
        "distutils2": "Distutils2",
        "django": "Django",
        "django_hstore": "amitu-hstore",
        "djangobower": "django-bower",
        "djcelery": "django-celery",
        "djkombu": "django-kombu",
        "djorm_pgarray": "djorm-ext-pgarray",
        "dns": "dnspython",
        "docgen": "ansible-docgenerator",
        "docker": "docker-py",
        "dogpile": "dogpile.core",
        "dogshell": "dogapi",
        "dot_parser": "pydot3k",
        "dotenv": "python-dotenv",
        "dpkt": "dpkt-fix",
        "ldif": "python-ldap",
        "durationfield": "django-durationfield",
        "dzclient": "datazilla",
        "easybuild": "easybuild-framework",
        "editor": "python-editor",
        "elasticluster": "azure-elasticluster-current",
        "elftools": "pyelftools",
        "elixir": "Elixir",
        "ell": "ell-ai",
        "emlib": "empy",
        "enchant": "pyenchant",
        "encutils": "cssutils",
        "engineio": "python-engineio",
        "enum": "enum34",
        "ephem": "pyephem",
        "errorreporter": "abl.errorreporter",
        "esplot": "beaker-es-plot",
        "example": "adrest",
        "examples": "tweepy",
        "ez_setup": "pycassa",
        "fabric": "Fabric",
        "faker": "Faker",
        "fedora": "python-fedora",
        "fias": "ailove-django-fias",
        "fiftyone_degrees": "51degrees-mobile-detector",
        "five": "five.pt",
        "flasher": "android-flasher",
        "flask": "Flask",
        "flask_frozen": "Frozen-Flask",
        "flask_redis": "Flask-And-Redis",
        "flaskext": "Flask-Bcrypt",
        "flvscreen": "vnc2flv",
        "followit": "django-followit",
        "forge": "pyforge",
        "formencode": "FormEncode",
        "formtools": "django-formtools",
        "fourch": "4ch",
        "franz": "allegrordf",
        "freetype": "freetype-py",
        "frontmatter": "python-frontmatter",
        "ftpcloudfs": "ftp-cloudfs",
        "funtests": "librabbitmq",
        "fuse": "fusepy",
        "fuzzy": "Fuzzy",
        "gabbi": "tiddlyweb",
        "gen_3dwallet": "3d-wallet-generator",
        "gendimen": "android-gendimen",
        "genshi": "Genshi",
        "quadtree": "python-geohash",
        "geonode": "GeoNode",
        "geoserver": "gsconfig",
        "geraldo": "Geraldo",
        "getenv": "django-getenv",
        "geventwebsocket": "gevent-websocket",
        "gflags": "python-gflags",
        "git": "GitPython",
        "github": "PyGithub",
        "github3": "github3.py",
        "gitpy": "git-py",
        "globusonline": "globusonline-transfer-api-client",
        "google": "protobuf",
        "grace-dizmo": "grace-dizmo",
        "grammar": "anovelmous-grammar",
        "grapheneapi": "graphenelib",
        "greplin": "scales",
        "grokcore": "grokcore.component",
        "gslib": "gsutil",
        "hamcrest": "PyHamcrest",
        "harpy": "HARPy",
        "hawk": "PyHawk-with-a-single-extra-commit",
        "haystack": "django-haystack",
        "hgext": "mercurial",
        "hggit": "hg-git",
        "hglib": "python-hglib",
        "sx": "pisa",
        "hola": "amarokHola",
        "hoover": "Hoover",
        "hostlist": "python-hostlist",
        "htmloutput": "nosehtmloutput",
        "hvad": "django-hvad",
        "hydra": "hydra-core",
        "ibis": "ibis-framework[duckdb]",
        "i99fix": "199Fix",
        "igraph": "python-igraph",
        "imdb": "IMDbPY",
        "impala": "impyla",
        "inmemorystorage": "ambition-inmemorystorage",
        "ipaddress": "backport-ipaddress",
        "jaraco": "jaraco.util",
        "jinja2": "Jinja2",
        "jiracli": "jira-cli",
        "johnny": "johnny-cache",
        "jpypex": "JPype1",
        "jsonfield": "django-jsonfield",
        "jstools": "aino-jstools",
        "jupyterpip": "jupyter-pip",
        "jwt": "PyJWT",
        "kazoo": "asana-kazoo",
        "kernprof": "line-profiler",
        "keyczar": "python-keyczar",
        "keyedcache": "django-keyedcache",
        "keystoneclient": "python-keystoneclient",
        "kickstarter": "kickstart",
        "krbv": "krbV",
        "kss": "kss.core",
        "kuyruk": "Kuyruk",
        "langconv": "AdvancedLangConv",
        "lava": "lava-utils-interface",
        "lazr": "lazr.uri",
        "ldaplib": "adpasswd",
        "lib2or3": "2or3",
        "lib3to2": "3to2",
        "libaito": "Aito",
        "libbe": "bugs-everywhere",
        "libbucket": "bucket",
        "libcloud": "apache-libcloud",
        "winreg": "future",
        "libgenerateDS": "generateDS",
        "libmproxy": "mitmproxy",
        "libsvm": "7lk-ocr-deploy",
        "lisa": "lisa-server",
        "loadingandsaving": "aspose-words-java-for-python",
        "locust": "locustio",
        "logbook": "Logbook",
        "logentries": "buildbot-status-logentries",
        "logilab": "logilab-mtconverter",
        "magic": "python-magic",
        "mako": "Mako",
        "manifestparser": "ManifestDestiny",
        "marionette": "marionette-client",
        "markdown": "Markdown",
        "marks": "pytest-marks",
        "markupsafe": "MarkupSafe",
        "mavnative": "pymavlink",
        "memcache": "python-memcached",
        "metacomm": "AllPairs",
        "metaphone": "Metafone",
        "metlog": "metlog-py",
        "mezzanine": "Mezzanine",
        "migrate": "sqlalchemy-migrate",
        "mimeparse": "python-mimeparse",
        "minitage": "minitage.recipe.common",
        "missingdrawables": "android-missingdrawables",
        "mkrst_themes": "2lazy2rest",
        "mockredis": "mockredispy",
        "modargs": "python-modargs",
        "model_utils": "django-model-utils",
        "models": "asposestorage",
        "moksha": "moksha.wsgi",
        "moneyed": "py-moneyed",
        "mongoalchemy": "MongoAlchemy",
        "monthdelta": "MonthDelta",
        "mopidy": "Mopidy",
        "mopytools": "MoPyTools",
        "mptt": "django-mptt",
        "mpv": "python-mpv",
        "mrbob": "mr.bob",
        "msgpack": "msgpack-python",
        "mutations": "aino-mutations",
        "mws": "amazon-mws",
        "mysql": "mysql-connector-repackaged",
        "native_tags": "django-native-tags",
        "ndg": "ndg-httpsclient",
        "nereid": "trytond-nereid",
        "nested": "baojinhuan",
        "nester": "bssm-pythonSig",
        "novaclient": "python-novaclient",
        "oauth2_provider": "alauda-django-oauth",
        "oauth2client": "oauth2client",
        "odf": "odfpy",
        "ometa": "Parsley",
        "openid": "python-openid",
        "opensearchsdk": "ali-opensearch",
        "oslo_i18n": "oslo.i18n",
        "oslo_serialization": "oslo.serialization",
        "oslo_utils": "oslo.utils",
        "oss": "aliyunoss",
        "output": "cashew",
        "owslib": "OWSLib",
        "rackdiag": "nwdiag",
        "paho": "paho-mqtt",
        "paintstore": "django-paintstore",
        "parler": "django-parler",
        "paste": "PasteScript",
        "path": "path.py",
        "patricia": "patricia-trie",
        "paver": "Paver",
        "peak": "ProxyTypes",
        "picasso": "anderson.picasso",
        "picklefield": "django-picklefield",
        "pivotal": "pivotal-py",
        "pwiz": "peewee",
        "plivoxml": "plivo",
        "plone": "plone.z3cform",
        "plonetheme": "plonetheme.barceloneta",
        "png": "pypng",
        "polymorphic": "django-polymorphic",
        "postmark": "python-postmark",
        "powerprompt": "bash-powerprompt",
        "prefetch": "django-prefetch",
        "printList": "AndrewList",
        "progressbar": "progressbar33",
        "provider": "django-oauth2-provider",
        "puresasl": "pure-sasl",
        "py7zlib": "pylzma",
        "pyAMI": "pyAMI-core",
        "pyarsespyder": "arsespyder",
        "pyasdf": "asdf",
        "pyaspell": "aspell-python-ctypes",
        "pybb": "pybbm",
        "pybloomfilter": "pybloomfiltermmap",
        "pyccuracy": "Pyccuracy",
        "pyck": "PyCK",
        "pycrfsuite": "python-crfsuite",
        "pydispatch": "PyDispatcher",
        "pygeolib": "pygeocoder",
        "pygments": "Pygments",
        "pygraph": "python-graph-core",
        "pyjon": "pyjon.utils",
        "pyjsonrpc": "python-jsonrpc",
        "pykka": "Pykka",
        "pylogo": "PyLogo",
        "pylons": "adhocracy-Pylons",
        "pymagic": "libmagic",
        "pymycraawler": "Amalwebcrawler",
        "pynma": "AbakaffeNotifier",
        "pyphen": "Pyphen",
        "pyrimaa": "AEI",
        "pysqlite2": "pysqlite",
        "pythongettext": "python-gettext",
        "pythonjsonlogger": "python-json-logger",
        "pyutilib": "PyUtilib",
        "pyximport": "Cython",
        "qs": "qserve",
        "quickapi": "django-quickapi",
        "quickunit": "nose-quickunit",
        "radical": "radical.utils",
        "readability": "readability-lxml",
        "readline": "gnureadline",
        "recaptcha_works": "django-recaptcha-works",
        "relstorage": "RelStorage",
        "reportapi": "django-reportapi",
        "requests": "Requests",
        "requirements": "requirements-parser",
        "rest_framework": "djangorestframework",
        "restclient": "py-restclient",
        "retrial": "async-retrial",
        "reversion": "django-reversion",
        "rhaptos2": "rhaptos2.common",
        "robot": "robotframework",
        "robots": "django-robots",
        "rosdep2": "rosdep",
        "rsbackends": "RSFile",
        "ruamel": "ruamel.base",
        "xmlenc": "pysaml2",
        "saga": "saga-python",
        "samtranslator": "aws-sam-translator",
        "sassutils": "libsass",
        "sayhi": "alex-sayhi",
        "scalrtools": "scalr",
        "scikits": "scikits.talkbox",
        "scratch": "scratchpy",
        "scss": "pyScss",
        "sdict": "dict.sorted",
        "sdk_updater": "android-sdk-updater",
        "sekizai": "django-sekizai",
        "sendfile": "pysendfile",
        "serial": "pyserial",
        "setuputils": "astor",
        "shapefile": "pyshp",
        "shapely": "Shapely",
        "sika": "ahonya-sika",
        "singleton": "pysingleton",
        "skbio": "scikit-bio",
        "skimage": "scikit-image",
        "sklearn": "scikit-learn",
        "skvideo": "scikit-video",
        "slack": "slackclient",
        "slugify": "unicode-slugify",
        "smarkets": "smk-python-sdk",
        "snappy": "ctypes-snappy",
        "snowflake": "snowflake-connector-python",
        "socketio": "gevent-socketio",
        "sockjs": "sockjs-tornado",
        "socks": "SocksiPy-branch",
        "solr": "solrpy",
        "solution": "Solution",
        "sorl": "sorl-thumbnail",
        "south": "South",
        "sphinx": "Sphinx",
        "sphinx_pypi_upload": "ATD-document",
        "sphinxcontrib": "sphinxcontrib-programoutput",
        "sqlalchemy": "SQLAlchemy",
        "src": "auto-mix-prep",
        "stats_toolkit": "bw-stats-toolkit",
        "statsd": "dogstatsd-python",
        "stdnum": "python-stdnum",
        "stoneagehtml": "StoneageHTML",
        "storages": "django-storages",
        "stubout": "mox",
        "suds": "suds-jurko",
        "swiftclient": "python-swiftclient",
        "test": "pytabix",
        "taggit": "django-taggit",
        "tastypie": "django-tastypie",
        "teamcity": "teamcity-messages",
        "telebot": "pyTelegramBotAPI",
        "tempita": "Tempita",
        "tenjin": "Tenjin",
        "termstyle": "python-termstyle",
        "thclient": "treeherder-client",
        "threaded_multihost": "django-threaded-multihost",
        "threecolor": "3color-Press",
        "tidylib": "pytidylib",
        "tlw": "3lwg",
        "toredis": "toredis-fork",
        "tornadoredis": "tornado-redis",
        "tower_cli": "ansible-tower-cli",
        "tracopt": "Trac",
        "translation_helper": "android-localization-helper",
        "treebeard": "django-treebeard",
        "trytond": "trytond-stock",
        "tsuru": "tsuru-circus",
        "tvrage": "python-tvrage",
        "tw2": "tw2.sqla",
        "twisted": "Twisted",
        "twitter": "python-twitter",
        "txclib": "transifex-client",
        "u115": "115wangpan",
        "umap": "umap-learn",
        "unidecode": "Unidecode",
        "universe": "ansible-universe",
        "usb": "pyusb",
        "useless": "useless.pipes",
        "userpass": "auth-userpass",
        "utilities": "automakesetup.py",
        "utkik": "aino-utkik",
        "uwsgidecorators": "uWSGI",
        "valentine": "ab",
        "validate": "configobj",
        "version": "chartio",
        "virtualenvapi": "ar-virtualenv-api",
        "vyatta": "brocade-plugins",
        "webob": "WebOb",
        "websocket": "websocket-client",
        "webtest": "WebTest",
        "werkzeug": "Werkzeug",
        "wheezy": "wheezy.http",
        "wikklytext": "tiddlywebwiki",
        "winrm": "pywinrm",
        "workflow": "Alfred-Workflow",
        "wsmeext": "WSME",
        "wtforms": "WTForms",
        "wtfpeewee": "wtf-peewee",
        "xdg": "pyxdg",
        "xdist": "pytest-xdist",
        "xmpp": "xmpppy",
        "xstatic": "XStatic-jquery-ui",
        "yaml": "PyYAML",
        "z3c": "z3c.zcmlhook",
        "zmq": "pyzmq",
        "zopyx": "zopyx.textindexng3",
    }



# Copyright 2024 Marimo. All rights reserved.
import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable

from marimo import _loggers
from marimo._messaging.ops import Interrupted
from marimo._runtime.context import get_context
from marimo._runtime.control_flow import MarimoInterrupt

LOGGER = _loggers.marimo_logger()

if TYPE_CHECKING:
    from marimo._runtime.runtime import Kernel


def construct_interrupt_handler(
    kernel: "Kernel",
) -> Callable[[int, Any], None]:
    def interrupt_handler(signum: int, frame: Any) -> None:
        """Tries to interrupt the kernel."""
        del signum
        del frame

        LOGGER.debug("Interrupt request received")
        # TODO(akshayka): if kernel is in `run` but not executing,
        # it won't be interrupted, which isn't right ... but the
        # probability of that happening is low.
        if kernel.execution_context is not None:
            Interrupted().broadcast()
            raise MarimoInterrupt

    return interrupt_handler


def construct_sigterm_handler(kernel: "Kernel") -> Callable[[int, Any], None]:
    del kernel

    @dataclass
    class Bit:
        value: bool = False

    shutting_down = Bit()

    def sigterm_handler(signum: int, frame: Any) -> None:
        """Cleans up the kernel and exits."""
        del signum
        del frame

        if shutting_down.value:
            # give previous SIGTERM a chance to quit ... makes
            # sure this method is reentrant
            return
        shutting_down.value = True

        get_context().virtual_file_registry.shutdown()
        # Force this process to exit.
        #
        # We use os._exit() instead of sys.exit() because we don't want the
        # child process to also run atexit handlers, which may result in
        # undefined behavior. Using sys.exit() on Linux sometimes causes
        # the parent process to hang on shutdown, leading to orphaned
        # processes and port.
        #
        # TODO(akshayka): The Python docs say this method is appropriate
        # for processes created with fork(), but they don't say anything
        # about processes made with spawn. macOS and Windows default to
        # spawn. If we have further issues with clean exits, we might
        # investigate here.
        #
        # https://docs.python.org/3/library/os.html#os._exit
        # https://www.unixguide.net/unix/programming/1.1.3.shtml
        os._exit(0)

    return sigterm_handler



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import io
import os
from typing import Optional, Union

import marimo._output.data.data as mo_data
from marimo._output.builder import h
from marimo._output.hypertext import Html
from marimo._output.rich_help import mddoc
from marimo._plugins.core.media import io_to_data_url


@mddoc
def audio(
    src: Union[str, io.BytesIO],
) -> Html:
    """Render an audio file as HTML.

    **Example.**

    ```python3
    mo.audio(
        src="https://upload.wikimedia.org/wikipedia/commons/8/8c/Ivan_Ili%C4%87-Chopin_-_Prelude_no._1_in_C_major.ogg"
    )

    mo.audio(src="path/to/local/file.wav")
    ```

    **Args.**

    - `src`: a path or URL to an audio file, bytes,
        or a file-like object opened in binary mode

    **Returns.**

    An audio player as an `Html` object.
    """
    resolved_src: Optional[str]

    if isinstance(src, (io.BufferedReader, io.BytesIO)):
        pos = src.tell()
        src.seek(0)
        resolved_src = mo_data.audio(src.read()).url
        src.seek(pos)
    elif isinstance(src, bytes):
        resolved_src = mo_data.audio(src).url
    elif isinstance(src, str) and os.path.isfile(os.path.expanduser(src)):
        src = os.path.expanduser(src)
        with open(src, "rb") as f:
            resolved_src = mo_data.audio(
                f.read(), ext=os.path.splitext(src)[1]
            ).url
    else:
        resolved_src = io_to_data_url(src, fallback_mime_type="audio/wav")

    return Html(h.audio(src=resolved_src, controls=True))



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Literal, Optional

from marimo._server.models.models import BaseResponse


@dataclass
class FileInfo:
    id: str
    path: str
    name: str
    is_directory: bool
    is_marimo_file: bool
    last_modified: Optional[float] = None
    children: List["FileInfo"] = field(default_factory=list)


@dataclass
class FileListRequest:
    # The directory path to list files from
    # If None, the root directory will be used
    path: Optional[str] = None


@dataclass
class FileDetailsRequest:
    # The path of the file or directory
    path: str


@dataclass
class FileOpenRequest:
    # The path of the file to open
    path: str


@dataclass
class FileTreeRequest:
    # The root directory path for the tree
    path: str


@dataclass
class FileCreateRequest:
    # The path where to create the file or directory
    path: str
    # 'file' or 'directory'
    type: Literal["file", "directory"]
    # The name of the file or directory
    name: str
    # The contents of the file, base64-encoded
    contents: Optional[str] = None


@dataclass
class FileDeleteRequest:
    # The path of the file or directory to delete
    path: str


@dataclass
class FileMoveRequest:
    # The current path of the file or directory
    path: str
    # The new path or name for the file or directory
    new_path: str


@dataclass
class FileUpdateRequest:
    # The current path of the file or directory
    path: str
    # The new contents of the file
    contents: str


@dataclass
class FileListResponse:
    files: List[FileInfo]
    root: str


@dataclass
class FileDetailsResponse:
    file: FileInfo
    contents: Optional[str] = None
    mime_type: Optional[str] = None


@dataclass
class FileCreateResponse(BaseResponse):
    # Additional information, e.g., error message
    message: Optional[str] = None
    info: Optional[FileInfo] = None


@dataclass
class FileDeleteResponse(BaseResponse):
    # Additional information, e.g., error message
    message: Optional[str] = None


@dataclass
class FileUpdateResponse(BaseResponse):
    # Additional information, e.g., error message
    message: Optional[str] = None
    info: Optional[FileInfo] = None


@dataclass
class FileMoveResponse(BaseResponse):
    # Additional information, e.g., error message
    message: Optional[str] = None
    info: Optional[FileInfo] = None



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from marimo._ast.cell import Cell
from marimo._messaging.mimetypes import KnownMimeType
from marimo._output import formatting
from marimo._output.formatters.formatter_factory import FormatterFactory


class CellFormatter(FormatterFactory):
    @staticmethod
    def package_name() -> None:
        return None

    def register(self) -> None:
        @formatting.formatter(Cell)
        def _format_cell(cell: Cell) -> tuple[KnownMimeType, str]:
            return cell._help()._mime_()



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from marimo._config.config import Theme
from marimo._messaging.mimetypes import KnownMimeType, MimeBundleOrTuple
from marimo._output.formatters.formatter_factory import FormatterFactory
from marimo._plugins.core.media import io_to_data_url

if TYPE_CHECKING:
    import altair  # type: ignore[import-not-found,import-untyped,unused-ignore] # noqa: E501


class AltairFormatter(FormatterFactory):
    @staticmethod
    def package_name() -> str:
        return "altair"

    def register(self) -> None:
        import altair  # type: ignore[import-not-found,import-untyped,unused-ignore] # noqa: E501

        from marimo._output import formatting
        from marimo._plugins.ui._impl.charts.altair_transformer import (
            register_transformers,
        )

        # add marimo transformers
        register_transformers()

        @formatting.formatter(altair.TopLevelMixin)
        def _show_chart(chart: altair.Chart) -> tuple[KnownMimeType, str]:
            import altair as alt

            # Try to get the _repr_mimebundle_ method from the chart
            # If its HTML, we want to handle this ourselves
            # if its svg, vega, or png, then we want to pass that instead
            # because that means the user has configured the that renderer
            mimebundle: MimeBundleOrTuple = {}
            try:
                mimebundle = chart._repr_mimebundle_() or {}  # type: ignore
            except Exception:
                pass

            # Handle where there are multiple mime types
            # return as a mimebundle
            if len(mimebundle) > 1:
                return (
                    "application/vnd.marimo+mimebundle",
                    json.dumps(mimebundle),
                )

            # Handle non-HTML mime types
            non_html_mime_types: list[KnownMimeType] = [
                "image/svg+xml",
                "image/png",
                "application/vnd.vega.v5+json",
                "application/vnd.vegalite.v5+json",
            ]
            for mime_type in non_html_mime_types:
                if mime_type in mimebundle:
                    mime_response = mimebundle[mime_type]
                    if isinstance(mime_response, bytes):
                        data_url = io_to_data_url(mime_response, mime_type)
                        return (mime_type, data_url or "")
                    if isinstance(mime_response, str):
                        return mime_type, mime_response
                    return mime_type, json.dumps(mime_response)

            # If vegafusion is enabled, just wrap in altair_chart
            if alt.data_transformers.active.startswith("vegafusion"):
                return (
                    "application/vnd.vega.v5+json",
                    chart.to_json(format="vega"),
                )

            # If the user has not set the max_rows option, we set it to 20_000
            # since we are able to handle the larger sizes (default is 5000)
            if "max_rows" not in alt.data_transformers.options:
                alt.data_transformers.options["max_rows"] = 20_000

            chart = _apply_embed_options(chart)

            # Return the chart as a vega-lite chart with embed options
            return ("application/vnd.vegalite.v5+json", chart.to_json())

    def apply_theme(self, theme: Theme) -> None:
        import altair as alt  # type: ignore

        alt.themes.enable("dark" if theme == "dark" else "default")  # type: ignore


# This is only needed since it seems that altair does not
# handle this internally.
# https://github.com/marimo-team/marimo/issues/2302
def _apply_embed_options(chart: altair.Chart) -> altair.Chart:
    import altair as alt

    # Respect user-set embed options
    # Note:
    # The python key is `embed_options`
    # The javascript key is `embedOptions`
    embed_options = alt.renderers.options.get("embed_options", {})
    prev_usermeta = {} if alt.Undefined is chart.usermeta else chart.usermeta
    chart["usermeta"] = {
        **prev_usermeta,
        "embedOptions": {
            **embed_options,
            **prev_usermeta.get("embedOptions", {}),
        },
    }
    return chart



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import asyncio
import base64
import dataclasses
import json
import signal
from typing import Any, Callable, Optional

from marimo import _loggers
from marimo._ast.cell import CellConfig, CellId_t
from marimo._config.config import MarimoConfig
from marimo._messaging.types import KernelMessage
from marimo._pyodide.streams import (
    PyodideStderr,
    PyodideStdin,
    PyodideStdout,
    PyodideStream,
)
from marimo._runtime import handlers, patches, requests
from marimo._runtime.context.kernel_context import initialize_kernel_context
from marimo._runtime.input_override import input_override
from marimo._runtime.marimo_pdb import MarimoPdb
from marimo._runtime.packages.pypi_package_manager import (
    MicropipPackageManager,
)
from marimo._runtime.requests import (
    AppMetadata,
    CodeCompletionRequest,
    ControlRequest,
    SetUIElementValueRequest,
)
from marimo._runtime.runtime import Kernel
from marimo._runtime.utils.set_ui_element_request_manager import (
    SetUIElementRequestManager,
)
from marimo._server.export.exporter import Exporter
from marimo._server.file_manager import AppFileManager
from marimo._server.files.os_file_system import OSFileSystem
from marimo._server.model import SessionMode
from marimo._server.models.base import deep_to_camel_case
from marimo._server.models.export import ExportAsHTMLRequest
from marimo._server.models.files import (
    FileCreateRequest,
    FileCreateResponse,
    FileDeleteRequest,
    FileDeleteResponse,
    FileDetailsRequest,
    FileListRequest,
    FileListResponse,
    FileMoveRequest,
    FileMoveResponse,
    FileUpdateRequest,
    FileUpdateResponse,
)
from marimo._server.models.models import (
    FormatRequest,
    FormatResponse,
    ReadCodeResponse,
    SaveAppConfigurationRequest,
    SaveNotebookRequest,
)
from marimo._server.session.session_view import SessionView
from marimo._snippets.snippets import read_snippets
from marimo._utils.formatter import DefaultFormatter
from marimo._utils.parse_dataclass import parse_raw

LOGGER = _loggers.marimo_logger()


class AsyncQueueManager:
    """Manages queues for a session."""

    def __init__(self) -> None:
        # Control messages for the kernel (run, set UI element, set config, etc
        # ) are sent through the control queue
        self.control_queue = asyncio.Queue[requests.ControlRequest]()

        # set UI elements duplicated in another queue so they can be batched
        self.set_ui_element_queue = asyncio.Queue[
            requests.SetUIElementValueRequest
        ]()

        # Code completion requests are sent through a separate queue
        self.completion_queue = asyncio.Queue[requests.CodeCompletionRequest]()

        # Input messages for the user's Python code are sent through the
        # input queue
        self.input_queue = asyncio.Queue[str](maxsize=1)

    def close_queues(self) -> None:
        # kernel thread cleans up read/write conn and IOloop handler on
        # exit; we don't join the thread because we don't want to block
        self.control_queue.put_nowait(requests.StopRequest())


class PyodideSession:
    """A client session that is compatible with Pyodide."""

    def __init__(
        self,
        app: AppFileManager,
        mode: SessionMode,
        on_write: Callable[[KernelMessage], None],
        app_metadata: AppMetadata,
        user_config: MarimoConfig,
    ) -> None:
        """Initialize kernel and client connection to it."""
        self.app_manager = app
        self.mode = mode
        self.app_metadata = app_metadata
        self._queue_manager = AsyncQueueManager()
        self.session_consumer = on_write
        self.session_view = SessionView()
        self._initial_user_config = user_config

        self.consumers: list[Callable[[KernelMessage], None]] = [
            lambda msg: self.session_consumer(msg),
            lambda msg: self.session_view.add_raw_operation(msg[1]),
        ]

    def _on_message(self, msg: KernelMessage) -> None:
        for consumer in self.consumers:
            consumer(msg)

    async def start(self) -> None:
        self.kernel_task = _launch_pyodide_kernel(
            control_queue=self._queue_manager.control_queue,
            set_ui_element_queue=self._queue_manager.set_ui_element_queue,
            completion_queue=self._queue_manager.completion_queue,
            input_queue=self._queue_manager.input_queue,
            on_message=self._on_message,
            is_edit_mode=self.mode == SessionMode.EDIT,
            configs=self.app_manager.app.cell_manager.config_map(),
            app_metadata=self.app_metadata,
            user_config=self._initial_user_config,
        )
        await self.kernel_task.start()

    def put_control_request(self, request: requests.ControlRequest) -> None:
        self._queue_manager.control_queue.put_nowait(request)
        if isinstance(request, requests.SetUIElementValueRequest):
            self._queue_manager.set_ui_element_queue.put_nowait(request)

    def put_completion_request(
        self, request: requests.CodeCompletionRequest
    ) -> None:
        self._queue_manager.completion_queue.put_nowait(request)

    def put_input(self, text: str) -> None:
        self._queue_manager.input_queue.put_nowait(text)

    def find_packages(self, code: str) -> list[str]:
        """
        Find the packages in the code based on the imports,
        and mapping from module names to package names.
        """
        import pyodide.code  # type: ignore

        imports: list[str] = pyodide.code.find_imports(code)  # type: ignore
        if not isinstance(imports, list):
            return []

        package_manager = MicropipPackageManager()
        return [package_manager.module_to_package(im) for im in imports]


class PyodideBridge:
    def __init__(
        self,
        session: PyodideSession,
    ):
        self.session = session
        self.file_system = OSFileSystem()

    def put_control_request(self, request: str) -> None:
        @dataclasses.dataclass
        class Container:
            body: requests.ControlRequest

        parsed = parse_raw({"body": json.loads(request)}, Container).body
        self.session.put_control_request(parsed)

    def put_input(self, text: str) -> None:
        self.session.put_input(text)

    def code_complete(self, request: str) -> None:
        parsed = parse_raw(json.loads(request), requests.CodeCompletionRequest)
        self.session.put_completion_request(parsed)

    def read_code(self) -> str:
        contents: str = self.session.app_manager.read_file()
        response = ReadCodeResponse(contents=contents)
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    async def read_snippets(self) -> str:
        snippets = await read_snippets()
        return json.dumps(deep_to_camel_case(dataclasses.asdict(snippets)))

    def format(self, request: str) -> str:
        parsed = parse_raw(json.loads(request), FormatRequest)
        formatter = DefaultFormatter(line_length=parsed.line_length)

        response = FormatResponse(codes=formatter.format(parsed.codes))
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def save(self, request: str) -> None:
        parsed = parse_raw(json.loads(request), SaveNotebookRequest)
        self.session.app_manager.save(parsed)

    def save_app_config(self, request: str) -> None:
        parsed = parse_raw(json.loads(request), SaveAppConfigurationRequest)
        self.session.app_manager.save_app_config(parsed.config)

    def save_user_config(self, request: str) -> None:
        parsed = parse_raw(json.loads(request), requests.SetUserConfigRequest)
        self.session.put_control_request(parsed)

    def rename_file(self, filename: str) -> None:
        self.session.app_manager.rename(filename)

    def list_files(
        self,
        request: str,
    ) -> str:
        body = parse_raw(json.loads(request), FileListRequest)
        root = body.path or self.file_system.get_root()
        files = self.file_system.list_files(root)
        response = FileListResponse(files=files, root=root)
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def file_details(
        self,
        request: str,
    ) -> str:
        body = parse_raw(json.loads(request), FileDetailsRequest)
        response = self.file_system.get_details(body.path)
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def create_file_or_directory(
        self,
        request: str,
    ) -> str:
        body = parse_raw(json.loads(request), FileCreateRequest)
        try:
            # If we need to eliminate the overhead associated with
            # base64-encoding/decoding the file contents, we could try pushing
            # filesystem operations into JavaScript
            decoded_contents = (
                base64.b64decode(body.contents)
                if body.contents is not None
                else None
            )
            info = self.file_system.create_file_or_directory(
                body.path, body.type, body.name, decoded_contents
            )
            response = FileCreateResponse(success=True, info=info)
        except Exception as e:
            response = FileCreateResponse(success=False, message=str(e))
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def delete_file_or_directory(
        self,
        request: str,
    ) -> str:
        body = parse_raw(json.loads(request), FileDeleteRequest)
        success = self.file_system.delete_file_or_directory(body.path)
        response = FileDeleteResponse(success=success)
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def move_file_or_directory(
        self,
        request: str,
    ) -> str:
        body = parse_raw(json.loads(request), FileMoveRequest)
        try:
            info = self.file_system.move_file_or_directory(
                body.path, body.new_path
            )
            response = FileMoveResponse(success=True, info=info)
        except Exception as e:
            response = FileMoveResponse(success=False, message=str(e))
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def update_file(
        self,
        request: str,
    ) -> str:
        body = parse_raw(json.loads(request), FileUpdateRequest)
        try:
            with open(body.path, "w") as file:
                file.write(body.contents)
            response = FileUpdateResponse(success=True)
        except Exception as e:
            response = FileUpdateResponse(success=False, message=str(e))
        return json.dumps(deep_to_camel_case(dataclasses.asdict(response)))

    def export_html(self, request: str) -> str:
        parsed = parse_raw(json.loads(request), ExportAsHTMLRequest)
        html, _filename = Exporter().export_as_html(
            file_manager=self.session.app_manager,
            session_view=self.session.session_view,
            display_config=self.session._initial_user_config["display"],
            request=parsed,
        )
        return json.dumps(html)

    def export_markdown(self, request: str) -> str:
        del request
        md, _filename = Exporter().export_as_md(
            file_manager=self.session.app_manager,
        )
        return json.dumps(md)


def _launch_pyodide_kernel(
    control_queue: asyncio.Queue[ControlRequest],
    set_ui_element_queue: asyncio.Queue[SetUIElementValueRequest],
    completion_queue: asyncio.Queue[CodeCompletionRequest],
    input_queue: asyncio.Queue[str],
    on_message: Callable[[KernelMessage], None],
    is_edit_mode: bool,
    configs: dict[CellId_t, CellConfig],
    app_metadata: AppMetadata,
    user_config: MarimoConfig,
) -> RestartableTask:
    from marimo._output.formatters.formatters import register_formatters

    register_formatters()

    LOGGER.debug("Launching kernel")

    # Patches for pyodide compatibility
    patches.patch_pyodide_networking()

    # Some libraries mess with Python's default recursion limit, which becomes
    # a problem when running with Pyodide.
    patches.patch_recursion_limit(limit=1000)

    # Create communication channels
    stream = PyodideStream(on_message, input_queue)
    stdout = PyodideStdout(stream)
    stderr = PyodideStderr(stream)
    stdin = PyodideStdin(stream) if is_edit_mode else None
    debugger = MarimoPdb(stdout=stdout, stdin=stdin) if is_edit_mode else None

    def _enqueue_control_request(req: ControlRequest) -> None:
        control_queue.put_nowait(req)
        if isinstance(req, SetUIElementValueRequest):
            set_ui_element_queue.put_nowait(req)

    kernel = Kernel(
        cell_configs=configs,
        app_metadata=app_metadata,
        stream=stream,
        stdout=stdout,
        stderr=stderr,
        stdin=stdin,
        module=patches.patch_main_module(
            file=app_metadata.filename, input_override=input_override
        ),
        enqueue_control_request=_enqueue_control_request,
        debugger_override=debugger,
        user_config=user_config,
    )
    initialize_kernel_context(
        kernel=kernel,
        stream=stream,
        stdout=stdout,
        stderr=stderr,
        virtual_files_supported=False,
        mode=SessionMode.EDIT if is_edit_mode else SessionMode.RUN,
    )

    if is_edit_mode:
        signal.signal(
            signal.SIGINT, handlers.construct_interrupt_handler(kernel)
        )

    ui_element_request_mgr = SetUIElementRequestManager(set_ui_element_queue)

    async def listen_messages() -> None:
        while True:
            request: ControlRequest | None = await control_queue.get()
            LOGGER.debug("received request %s", request)
            if isinstance(request, requests.SetUIElementValueRequest):
                request = ui_element_request_mgr.process_request(request)

            if request is not None:
                await kernel.handle_message(request)

    async def listen_completion() -> None:
        while True:
            request = await completion_queue.get()
            while not completion_queue.empty():
                # discard stale requests to avoid choking the runtime
                request = await completion_queue.get()
            LOGGER.debug("received completion request %s", request)
            # 5 is arbitrary, but is a good limit:
            # too high will cause long load times
            # too low can be not as useful
            kernel.code_completion(request, docstrings_limit=5)

    async def listen() -> None:
        await asyncio.gather(listen_messages(), listen_completion())

    return RestartableTask(listen)


class RestartableTask:
    def __init__(self, coro: Callable[[], Any]):
        self.coro = coro
        self.task: Optional[asyncio.Task[Any]] = None
        self.stopped = False

    async def start(self) -> None:
        """Create a task that runs the coro."""
        while True:
            if self.stopped:
                break

            try:
                self.task = asyncio.create_task(self.coro())
                await self.task
            except asyncio.CancelledError:
                pass

    def stop(self) -> None:
        # Stop the task and set the stopped flag
        self.stopped = True
        assert self.task is not None
        self.task.cancel()

    def restart(self) -> None:
        # Cancel the current task, which will cause
        # the while loop to start a new task
        assert self.task is not None
        self.task.cancel()



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from typing import Any, Dict, Literal, Tuple, Union

# It is convenient to write mimetypes as strings,
# but can lead to typos. This literal type
# helps us avoid typos.
KnownMimeType = Literal[
    "application/json",
    "application/vnd.marimo+error",
    "application/vnd.marimo+traceback",
    "application/vnd.marimo+mimebundle",
    "application/vnd.vega.v5+json",
    "application/vnd.vegalite.v5+json",
    "image/png",
    "image/svg+xml",
    "image/tiff",
    "image/avif",
    "image/bmp",
    "image/gif",
    "image/jpeg",
    "video/mp4",
    "video/mpeg",
    "text/html",
    "text/plain",
    "text/markdown",
    "text/latex",
    "text/csv",
]

MimeBundle = Dict[KnownMimeType, Any]
MimeBundleOrTuple = Union[MimeBundle, Tuple[MimeBundle, Any]]



# Copyright 2024 Marimo. All rights reserved.

from __future__ import annotations

from typing import Any

from marimo._config.settings import GLOBAL_SETTINGS

# Print helpers


def bold(text: str) -> str:
    return "\033[1m" + text + "\033[0m"


def green(text: str, bold: bool = False) -> str:
    prefix = "\033[32m" if not bold else "\033[1;32m"
    return prefix + text + "\033[0m"


def orange(text: str, bold: bool = False) -> str:
    prefix = "\033[33m" if not bold else "\033[1;33m"
    return prefix + text + "\033[0m"


def red(text: str, bold: bool = False) -> str:
    prefix = "\033[31m" if not bold else "\033[1;31m"
    return prefix + text + "\033[0m"


def muted(text: str) -> str:
    return "\033[90m" + text + "\033[0m"


def echo(*args: Any, **kwargs: Any) -> None:
    import click

    if GLOBAL_SETTINGS.QUIET:
        return

    click.echo(*args, **kwargs)



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from typing import (
    Dict,
    List,
    Tuple,
    TypeVar,
    Union,
)

from narwhals.typing import IntoDataFrame

from marimo import _loggers
from marimo._output.mime import MIME
from marimo._plugins.core.web_component import JSONType

LOGGER = _loggers.marimo_logger()

T = TypeVar("T")
Numeric = Union[int, float]
ListOrTuple = Union[List[T], Tuple[T, ...]]


TableData = Union[
    List[JSONType],
    ListOrTuple[Union[str, int, float, bool, MIME, None]],
    ListOrTuple[Dict[str, JSONType]],
    Dict[str, ListOrTuple[JSONType]],
    IntoDataFrame,
]



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import json
from textwrap import dedent
from typing import TYPE_CHECKING, List, Optional, Union, cast

from marimo import __version__, _loggers
from marimo._ast.app import App, InternalApp, _AppConfig
from marimo._ast.cell import Cell, CellConfig
from marimo._ast.compiler import compile_cell
from marimo._messaging.cell_output import CellOutput
from marimo._output.formatting import as_html, mime_to_html
from marimo._output.utils import uri_encode_component
from marimo._plugins.ui import code_editor
from marimo._server.export import run_app_until_completion
from marimo._server.file_manager import AppFileManager
from marimo._server.file_router import AppFileRouter
from marimo._utils.marimo_path import MarimoPath

if TYPE_CHECKING:
    from marimo._server.session.session_view import SessionView

LOGGER = _loggers.marimo_logger()


class MarimoIslandStub:
    def __init__(
        self,
        display_code: bool = False,
        display_output: bool = True,
        is_reactive: bool = True,
        *,
        cell_id: str,
        app_id: str,
        code: str,
    ):
        self._cell_id = cell_id
        self._app_id = app_id
        self._code = code
        self._display_code = display_code
        self._display_output = display_output
        self._is_reactive = is_reactive

        self._internal_app: Optional[InternalApp] = None
        self._session_view: Optional[SessionView] = None
        self._output: Optional[CellOutput] = None

    @property
    def output(self) -> Optional[CellOutput]:
        # Leave output accessible for direct use for non-interactive cases e.g.
        # pdf.
        if self._output is None:
            if self._session_view is not None:
                outputs = self._session_view.get_cell_outputs([self._cell_id])
                self._output = outputs.get(self._cell_id, None)
        return self._output

    @property
    def code(self) -> str:
        return self._code

    def render(
        self,
        display_code: Optional[bool] = None,
        display_output: Optional[bool] = None,
        is_reactive: Optional[bool] = None,
    ) -> str:
        """
        Render the HTML island code for the cell.
        Note: This will override construction defaults.

        *Args:*

        - display_code (bool): Whether to display the code in HTML.
        - display_output (bool): Whether to include the output in the HTML.
        - is_reactive (bool): Whether this code block will run with pyodide.

        *Returns:*

        - str: The HTML code.
        """

        is_reactive = (
            is_reactive if is_reactive is not None else self._is_reactive
        )
        display_code = (
            display_code if display_code is not None else self._display_code
        )
        display_output = (
            display_output
            if display_output is not None
            else self._display_output
        )

        if not (display_code or display_output or is_reactive):
            raise ValueError("You must include either code or output")

        output = (
            mime_to_html(self.output.mimetype, self.output.data)
            if self.output is not None
            else None
        )

        # Specifying display_code=False will hide the code block, but still
        # make it present for reactivity, unless reactivity is disabled.
        if display_code:
            # TODO: Allow for non-disabled code editors.
            code_block = as_html(
                code_editor(self.code.strip(), disabled=False)
            ).text
        else:
            code_block = (
                "<marimo-cell-code hidden>"
                f"{uri_encode_component(self.code) if is_reactive else ''}"
                "</marimo-cell-code>"
            )

        # Cell may not have output
        # (e.g. imports, but still needs to be included)
        return remove_empty_lines(
            dedent(
                f"""
        <marimo-island
            data-app-id="{self._app_id}"
            data-cell-id="{self._cell_id}"
            data-reactive="{json.dumps(is_reactive)}"
        >
            <marimo-cell-output>
            {output.text if output and display_output else ""}
            </marimo-cell-output>
            {code_block}
        </marimo-island>
        """
            )
        ).strip()


class MarimoIslandGenerator:
    """
    Generates Marimo islands for embedding in other pages.

    This is a great way to use another SSG framework that converts
    Python code to HTML using marimo-islands.

    Generally you will want to:

    1. Find all the code snippets and add them to the generator.
    2. Build the app.
    3. Replace all code snippets with the rendered HTML.
    4. Include the header in the <head> tag.

    # Example

    ```python
    from marimo.islands import MarimoIslandGenerator

    generator = MarimoIslandGenerator()
    block1 = generator.add_code("import marimo as mo")
    block2 = generator.add_code("mo.md('Hello, islands!')")

    # Build the app
    app = await generator.build()

    # Render the app
    output = f\"\"\"
    <html>
        <head>
            {generator.render_head()}
        </head>
        <body>
            {block1.render(display_output=False)}
            {block2.render()}
        </body>
    </html>
    \"\"\"
    ```
    """

    def __init__(self, app_id: str = "main"):
        self.has_run = False
        self._app_id = app_id
        self._app = InternalApp(App())
        self._stubs: List[MarimoIslandStub] = []
        self._config = _AppConfig()

    @staticmethod
    def from_file(
        filename: str,
        display_code: bool = False,
    ) -> MarimoIslandGenerator:
        """
        Create a MarimoIslandGenerator and populate MarimoIslandStubs
        using code cells from a marimo *.py file.

        *Args:*

        - filename (str): Marimo .py filename to convert to reactive HTML.
        - display_code (bool): Whether to display the code in HTML snippets.
        """
        path = MarimoPath(filename)
        file_router = AppFileRouter.from_filename(path)
        file_key = file_router.get_unique_file_key()
        assert file_key is not None
        file_manager = file_router.get_file_manager(file_key)

        generator = MarimoIslandGenerator()
        stubs = []
        for cell_data in file_manager.app.cell_manager.cell_data():
            stubs.append(
                generator.add_code(
                    cell_data.code,
                    display_code=display_code,
                )
            )

        generator._stubs = stubs
        generator._config = file_manager.app.config

        return generator

    def add_code(
        self,
        code: str,
        display_code: bool = False,
        display_output: bool = True,
        is_reactive: bool = True,
        is_raw: bool = False,
    ) -> MarimoIslandStub:
        """Add a code cell to the app.

        *Args:*

        - code (str): The code to add to the app.
        - display_code (bool): Whether to display the code in the HTML.
        - display_output (bool): Whether to display the output in the HTML.
        - is_raw (bool): Whether to handled the code without formatting.
        - is_reactive (bool): Whether this code block will run with pyodide.
        """
        if not is_raw:
            code = dedent(code)

        cell_id = self._app.cell_manager.create_cell_id()
        cell_impl = compile_cell(code, cell_id)
        cell_impl.configure(CellConfig(hide_code=False))
        cell = Cell(_name="__", _cell=cell_impl)

        self._app.cell_manager._register_cell(
            cell,
            app=self._app,
        )

        stub = MarimoIslandStub(
            cell_id=cell_id,
            app_id=self._app_id,
            code=code,
            display_code=display_code,
            display_output=display_output,
            is_reactive=is_reactive,
        )
        self._stubs.append(stub)

        return stub

    async def build(self) -> App:
        """
        Build the app. This should be called after adding all the code cells.

        *Returns:*

        - App: The built app.
        """
        if self.has_run:
            raise ValueError("You can only call build() once")

        session = await run_app_until_completion(
            file_manager=AppFileManager.from_app(self._app),
            cli_args={},
        )
        self.has_run = True

        for stub in self._stubs:
            stub._internal_app = self._app
            stub._session_view = session

        return cast(App, self._app)

    def render_head(
        self,
        *,
        version_override: str = __version__,
        _development_url: Union[str, bool] = False,
    ) -> str:
        """
        Render the header for the app.
        This should be included in the <head> tag of the page.

        *Args:*

        - version_override (str): Marimo version to use for loaded js/css.
        - _development_url (str): If True, uses local marimo islands js.
        """

        # This loads:
        # - The marimo islands js
        # - The marimo islands css
        # - Preconnects to Google Fonts (https://stackoverflow.com/questions/73838138)
        # - Fonts from Google Fonts
        #   (otherwise they would get bundled in the css)
        # - Fonts from KaTeX
        #   (otherwise they would get bundled in the css)

        base_url = f"https://cdn.jsdelivr.net/npm/@marimo-team/islands@{version_override}"
        # This should be kept in sync fonts.css in the frontend
        # Since this is embedded on other pages, we want display=swap
        # for the most compatible font loading
        font_url = "https://fonts.googleapis.com/css2?family=Fira+Mono:wght@400;500;700&amp;family=Lora&amp;family=PT+Sans:wght@400;700&amp;display=swap"

        fonts = f"""
            <link rel="preconnect" href="https://fonts.googleapis.com" />
            <link
                rel="preconnect"
                href="https://fonts.gstatic.com"
                crossorigin
            />
            <link href="{font_url}" rel="stylesheet" />
            <link
                rel="stylesheet"
                href="https://cdn.jsdelivr.net/npm/katex@0.16.10/dist/katex.min.css"
                integrity="sha384-wcIxkf4k558AjM3Yz3BBFQUbk/zgIYC2R0QpeeYb+TwlBVMrlgLqwRjRtGZiK7ww"
                crossorigin="anonymous"
            />
        """.strip()

        if _development_url:
            base_url = "http://localhost:5174"
            if isinstance(_development_url, str):
                base_url = _development_url
            return dedent(
                f"""
                <script
                    type="module"
                    src="{base_url}/src/core/islands/main.ts"
                ></script>
                {fonts}
                """
            ).strip()

        marimo_tags = """
        <marimo-filename hidden></marimo-filename>
        <marimo-mode data-mode='read' hidden></marimo-mode>
        """.strip()

        return dedent(
            f"""
            <script type="module" src="{base_url}/dist/main.js"></script>
            <link
                href="{base_url}/dist/style.css"
                rel="stylesheet"
                crossorigin="anonymous"
            />
            {fonts}
            {marimo_tags}
            """
        ).strip()

    def render_init_island(self) -> str:
        """
        Renders a static html MarimoIsland str which displays a spinning
        initialization loader while Pyodide loads and disappears once
        the kernel is ready to use.
        """

        init_cell_id = self._app.cell_manager.create_cell_id()
        init_output = """
        <div class="marimo">
          <div class="flex flex-col flex-1 items-center justify-center">
            <svg
              xmlns="http://www.w3.org/2000/svg"
              width="24"
              height="24"
              viewBox="0 0 24 24"
              fill="none"
              stroke="currentColor"
              stroke-width="1"
              stroke-linecap="round"
              stroke-linejoin="round"
              class="size-20 animate-spin text-primary"
            >
              <path d="M21 12a9 9 0 1 1-6.219-8.56"></path>
            </svg>
            <div>Initializing...</div>
          </div>
        </div>
        """
        init_island = dedent(
            f"""
            <marimo-island
                data-app-id="{self._app_id}"
                data-cell-id="{init_cell_id}"
                data-reactive="{json.dumps(False)}"
            >
                <marimo-cell-output>
                {init_output}
                </marimo-cell-output>
                <marimo-cell-code hidden></marimo-cell-code>
            </marimo-island>
            """
        ).strip()

        return init_island

    def render_body(
        self,
        *,
        include_init_island: bool = True,
        max_width: Optional[str] = None,
        margin: Optional[str] = None,
        style: Optional[str] = None,
    ) -> str:
        """
        Render the body for the app.
        This should be included in the <body> tag of the page.

        *Args:*
        - include_init_island (bool): If True, adds initialization loader.
        - max_width (str): CSS style max_width property.
        - margin (str): CSS style margin property.
        - style (str): CSS style. Overrides max_width and margin.
        """

        rendered_stubs = []
        for stub in self._stubs:
            rendered_stubs.append(stub.render())

        if include_init_island:
            init_island = self.render_init_island()
            rendered_stubs = [init_island] + rendered_stubs

        body = "\n".join(rendered_stubs)

        if margin is None:
            margin = "auto"
        if max_width is None:
            width = self._config.width
            if width == "compact" or width == "normal":
                max_width = "740px"
            elif width == "medium":
                max_width = "1110px"
            else:
                max_width = "none"

        if style is None:
            style = f"margin: {margin}; max-width: {max_width};"

        return dedent(
            f"""
                <div style="{style}">
                  {body}
                </div>
                """
        ).strip()

    def render_html(
        self,
        *,
        version_override: str = __version__,
        _development_url: Union[str, bool] = False,
        include_init_island: bool = True,
        max_width: Optional[str] = None,
        margin: Optional[str] = None,
        style: Optional[str] = None,
    ) -> str:
        """
        Render reactive html for the app.

        *Args:*

        - version_override (str): Marimo version to use for loaded js/css.
        - _development_url (str): If True, uses local marimo islands js.
        - include_init_island (bool): If True, adds initialization loader.
        - max_width (str): CSS style max_width property.
        - margin (str): CSS style margin property.
        - style (str): CSS style. Overrides max_width and margin.
        """
        head = self.render_head(
            version_override=version_override,
            _development_url=_development_url,
        )
        body = self.render_body(
            include_init_island=include_init_island,
            max_width=max_width,
            margin=margin,
            style=style,
        )
        title = (
            self._app_id
            if self._config.app_title is None
            else self._config.app_title
        )

        return dedent(
            f"""<!doctype html>
                <html lang="en">
                    <head>
                      <meta charset="UTF-8" />
                      <title> {title} </title>
                        {head}
                    </head>
                    <body>
                      {body}
                    </body>
                </html>
                """
        ).strip()


def remove_empty_lines(text: str) -> str:
    return "\n".join([line for line in text.split("\n") if line.strip() != ""])



# Copyright 2024 Marimo. All rights reserved.
import marimo

__generated_with = "0.3.9"
app = marimo.App()


@app.cell
def __(mo):
    mo.md(
        r"""
        # Pandas DataFrame: Query by regexp (regular expression)
        """
    )
    return


@app.cell
def __():
    import pandas as pd

    df = pd.DataFrame(
        {
            "first_name": ["Sarah", "John", "Kyle", "Joe"],
            "last_name": ["Connor", "Connor", "Reese", "Bonnot"],
        }
    )

    df[df.last_name.str.match(".*onno.*")]
    return df, pd


@app.cell
def __():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()



# Copyright 2024 Marimo. All rights reserved.

import marimo

__generated_with = "0.4.0"
app = marimo.App()


@app.cell
def __(mo):
    mo.md(
        r"""
        # Query Parameters: Reading query parameters

        Use `mo.query_params` to access query parameters passed to the notebook.
        """
    )
    return


@app.cell
def __(mo):
    params = mo.query_params()
    print(params)
    return (params,)


@app.cell
def __():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()



# Copyright 2024 Marimo. All rights reserved.
class CycleError(Exception):
    pass


class MultipleDefinitionError(Exception):
    pass


class DeleteNonlocalError(Exception):
    pass


class UnparsableError(Exception):
    pass



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import abc
import json
import logging
import os
import pathlib
import urllib.parse
import urllib.request
from tempfile import TemporaryDirectory
from typing import Optional, Tuple
from urllib.error import HTTPError

from marimo._cli.print import green
from marimo._utils.marimo_path import MarimoPath
from marimo._utils.url import is_url


def is_github_src(url: str, ext: str) -> bool:
    if not is_url(url):
        return False

    hostname = urllib.parse.urlparse(url).hostname
    if hostname != "github.com" and hostname != "raw.githubusercontent.com":
        return False
    path: str = urllib.parse.urlparse(url).path
    if not path.endswith(ext):
        return False
    return True


def get_github_src_url(url: str) -> str:
    # Change hostname to raw.githubusercontent.com
    path = urllib.parse.urlparse(url).path
    path = path.replace("/blob/", "/", 1)
    return f"https://raw.githubusercontent.com{path}"


class FileReader(abc.ABC):
    @abc.abstractmethod
    def can_read(self, name: str) -> bool:
        pass

    @abc.abstractmethod
    def read(self, name: str) -> Tuple[str, str]:
        """Read the file and return its content and filename."""
        pass


class LocalFileReader(FileReader):
    def can_read(self, name: str) -> bool:
        return not is_url(name)

    def read(self, name: str) -> Tuple[str, str]:
        # Is directory
        if os.path.isdir(name):
            return "", os.path.basename(name)
        with open(name, "r") as f:
            content = f.read()
        return content, os.path.basename(name)


class GitHubIssueReader(FileReader):
    def can_read(self, name: str) -> bool:
        return is_url(name) and name.startswith(
            "https://github.com/marimo-team/marimo/issues/"
        )

    def read(self, name: str) -> Tuple[str, str]:
        issue_number = name.split("/")[-1]
        api_url = f"https://api.github.com/repos/marimo-team/marimo/issues/{issue_number}"
        issue_response = urllib.request.urlopen(api_url).read().decode("utf-8")
        issue_json = json.loads(issue_response)
        body = issue_json["body"]
        code = self._find_python_code_in_github_issue(body)
        return code, f"issue_{issue_number}.py"

    @staticmethod
    def _find_python_code_in_github_issue(body: str) -> str:
        return body.split("```python")[1].rsplit("```", 1)[0]


class StaticNotebookReader(FileReader):
    CODE_PREFIX = '<marimo-code hidden="">'
    CODE_SUFFIX = "</marimo-code>"
    FILENAME_PREFIX = '<marimo-filename hidden="">'
    FILENAME_SUFFIX = "</marimo-filename>"

    def can_read(self, name: str) -> bool:
        return self._is_static_marimo_notebook_url(name)[0]

    def read(self, name: str) -> Tuple[str, str]:
        _, file_contents = self._is_static_marimo_notebook_url(name)
        code = self._extract_code_from_static_notebook(file_contents)
        filename = self._extract_filename_from_static_notebook(file_contents)
        return code, filename

    @staticmethod
    def _is_static_marimo_notebook_url(url: str) -> tuple[bool, str]:
        def download(url: str) -> tuple[bool, str]:
            logging.info("Downloading %s", url)
            request = urllib.request.Request(
                url,
                # User agent to avoid 403 Forbidden some bot protection
                headers={"User-Agent": "Mozilla/5.0"},
            )
            file_contents = (
                urllib.request.urlopen(request).read().decode("utf-8")
            )
            return StaticNotebookReader.CODE_PREFIX in file_contents, str(
                file_contents
            )

        # Not a URL
        if not is_url(url):
            return False, ""

        # Ends with .html, try to download it
        if url.endswith(".html"):
            return download(url)

        # Starts with https://static.marimo.app/, append /download
        if url.startswith("https://static.marimo.app/"):
            return download(os.path.join(url, "download"))

        # Otherwise, not a static marimo notebook
        return False, ""

    @staticmethod
    def _extract_code_from_static_notebook(file_contents: str) -> str:
        # normalize hidden attribute
        file_contents = file_contents.replace("hidden=''", 'hidden=""')
        return file_contents.split(StaticNotebookReader.CODE_PREFIX)[1].split(
            StaticNotebookReader.CODE_SUFFIX
        )[0]

    @staticmethod
    def _extract_filename_from_static_notebook(file_contents: str) -> str:
        # normalize hidden attribute
        file_contents = file_contents.replace("hidden=''", 'hidden=""')
        return file_contents.split(StaticNotebookReader.FILENAME_PREFIX)[
            1
        ].split(StaticNotebookReader.FILENAME_SUFFIX)[0]


class GitHubSourceReader(FileReader):
    def can_read(self, name: str) -> bool:
        return is_github_src(name, ext=".py") or is_github_src(name, ext=".md")

    def read(self, name: str) -> Tuple[str, str]:
        url = get_github_src_url(name)
        content = urllib.request.urlopen(url).read().decode("utf-8")
        return content, os.path.basename(url)


class GenericURLReader(FileReader):
    def can_read(self, name: str) -> bool:
        return is_url(name)

    def read(self, name: str) -> Tuple[str, str]:
        content = urllib.request.urlopen(name).read().decode("utf-8")
        # Remove query parameters from the URL
        url_without_query = name.split("?")[0]
        return content, os.path.basename(url_without_query)


class FileContentReader:
    def __init__(self) -> None:
        self.readers = [
            LocalFileReader(),
            GitHubIssueReader(),
            StaticNotebookReader(),
            GitHubSourceReader(),
            GenericURLReader(),
        ]

    def read_file(self, name: str) -> Tuple[str, str]:
        """
        Read the file and return its content and filename

        Args:
            name (str): File path or URL

        Raises:
            ValueError: If the file cannot be read

        Returns:
            Tuple[str, str]: File content and filename
        """
        for reader in self.readers:
            if reader.can_read(name):
                return reader.read(name)
        raise ValueError(f"Unable to read file contents of {name}")


class FileHandler(abc.ABC):
    @abc.abstractmethod
    def can_handle(self, name: str) -> bool:
        pass

    @abc.abstractmethod
    def handle(
        self, name: str, temp_dir: TemporaryDirectory[str]
    ) -> Tuple[str, Optional[TemporaryDirectory[str]]]:
        pass


class LocalFileHandler(FileHandler):
    def __init__(self, allow_new_file: bool, allow_directory: bool):
        self.allow_new_file = allow_new_file
        self.allow_directory = allow_directory

    def can_handle(self, name: str) -> bool:
        return not is_url(name)

    def handle(
        self, name: str, temp_dir: TemporaryDirectory[str]
    ) -> Tuple[str, Optional[TemporaryDirectory[str]]]:
        del temp_dir
        import click

        path = pathlib.Path(name)

        if self.allow_directory and os.path.isdir(name):
            return name, None

        if path.suffix == ".ipynb":
            prefix = str(path)[: -len(".ipynb")]
            raise click.ClickException(
                f"Invalid NAME - {name} is not a Python file.\n\n"
                f"  {green('Tip:')} Convert {name} to a marimo notebook with"
                "\n\n"
                f"    marimo convert {name} > {prefix}.py\n\n"
                f"  then open with marimo edit {prefix}.py"
            )

        if not MarimoPath.is_valid_path(path):
            raise click.ClickException(
                f"Invalid NAME - {name} is not a Python or Markdown file"
            )

        if not self.allow_new_file:
            if not os.path.exists(name):
                raise click.ClickException(
                    f"Invalid NAME - {name} does not exist"
                )
            if not path.is_file():
                raise click.ClickException(
                    f"Invalid NAME - {name} is not a file"
                )

        return name, None


class RemoteFileHandler(FileHandler):
    def __init__(self) -> None:
        self.reader = FileContentReader()

    def can_handle(self, name: str) -> bool:
        return is_url(name)

    def handle(
        self, name: str, temp_dir: TemporaryDirectory[str]
    ) -> Tuple[str, Optional[TemporaryDirectory[str]]]:
        try:
            content, filename = self.reader.read_file(name)
        except HTTPError as e:
            import click

            raise click.ClickException(f"Failed to read URL: {e}")  # noqa: B904
        path_to_app = self._create_tmp_file_from_content(
            content, filename, temp_dir
        )
        return path_to_app, temp_dir

    @staticmethod
    def _create_tmp_file_from_content(
        content: str, name: str, temp_dir: TemporaryDirectory[str]
    ) -> str:
        logging.info("Creating temporary file")
        path_to_app = os.path.join(temp_dir.name, name)
        with open(path_to_app, "w") as f:
            f.write(content)
        logging.info("App saved to %s", path_to_app)
        return path_to_app


def validate_name(
    name: str, allow_new_file: bool, allow_directory: bool
) -> Tuple[str, Optional[TemporaryDirectory[str]]]:
    """
    Validate the file name and return the path to the file.

    Args:
        name (str): Local file path, URL, or directory path
        allow_new_file (bool): Whether to allow creating a new file
        allow_directory (bool): Whether to allow a directory

    Raises:
        ValueError: If the file name is invalid

    Returns:
        Path to the file and temporary directory
    """
    handlers = [
        LocalFileHandler(allow_new_file, allow_directory),
        RemoteFileHandler(),
    ]

    temp_dir = TemporaryDirectory()

    for handler in handlers:
        if handler.can_handle(name):
            return handler.handle(name, temp_dir)

    raise ValueError(f"Unable to handle file {name}")



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import ast
import datetime
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Final,
    List,
    Optional,
    TypedDict,
    Union,
    cast,
)

import narwhals.stable.v1 as nw
from narwhals.typing import IntoDataFrame

import marimo._output.data.data as mo_data
from marimo import _loggers
from marimo._output.rich_help import mddoc
from marimo._plugins.ui._core.ui_element import UIElement
from marimo._plugins.ui._impl.tables.utils import get_table_manager

LOGGER = _loggers.marimo_logger()

if TYPE_CHECKING:
    from narwhals.dtypes import DType


@dataclass
class DataEditorValue:
    # Row-oriented data
    data: List[Dict[str, Any]]


class PositionalEdit(TypedDict):
    rowIdx: int
    columnId: str
    value: Any


class DataEdits(TypedDict):
    edits: List[PositionalEdit]


RowOrientedData = List[Dict[str, Any]]
ColumnOrientedData = Dict[str, List[Any]]


@mddoc
class data_editor(
    UIElement[
        DataEdits,
        Union[RowOrientedData, ColumnOrientedData, IntoDataFrame],
    ]
):
    """
    **[EXPERIMENTAL]**

    This component is experimental and intentionally limited in features,
    if you have any feature requests, please file an issue at
    https://github.com/marimo-team/marimo/issues.

    A data editor component for editing tabular data.

    The data can be supplied as:
    1. a Pandas, Polars, or Pyarrow DataFrame
    2. a list of dicts, with one dict for each row, keyed by column names
    3. a dict of lists, with each list representing a column

    **Examples.**

    Create a data editor from a Pandas dataframe:

    ```python
    import pandas as pd

    df = pd.DataFrame({"A": [1, 2, 3], "B": ["a", "b", "c"]})
    editor = mo.ui.experimental_data_editor(data=df, label="Edit Data")
    ```

    Create a data editor from a list of dicts:

    ```python
    data = [{"A": 1, "B": "a"}, {"A": 2, "B": "b"}, {"A": 3, "B": "c"}]
    editor = mo.ui.experimental_data_editor(data=data, label="Edit Data")
    ```

    Create a data editor from a dict of lists:

    ```python
    data = {"A": [1, 2, 3], "B": ["a", "b", "c"]}
    editor = mo.ui.experimental_data_editor(data=data, label="Edit Data")
    ```

    **Attributes.**

    - `value`: the current state of the edited data
    - `data`: the original data passed to the editor

    **Initialization Args.**

    - `data`: The data to be edited. Can be a Pandas dataframe,
        a list of dicts, or a dict of lists.
    - `label`: markdown label for the element
    - `on_change`: optional callback to run when this element's value changes
    """

    _name: Final[str] = "marimo-data-editor"

    LIMIT: Final[int] = 1000

    def __init__(
        self,
        data: Union[RowOrientedData, ColumnOrientedData, IntoDataFrame],
        *,
        pagination: bool = True,
        page_size: int = 50,
        label: str = "",
        on_change: Optional[
            Callable[
                [Union[RowOrientedData, ColumnOrientedData, IntoDataFrame]],
                None,
            ]
        ] = None,
    ) -> None:
        table_manager = get_table_manager(data)

        size = table_manager.get_num_rows()
        if size is None or size > self.LIMIT:
            raise ValueError(
                f"Data editor supports a maximum of {self.LIMIT} rows."
            )

        self._data = data
        self._edits: DataEdits | None = None
        field_types = table_manager.get_field_types()

        super().__init__(
            component_name=data_editor._name,
            label=label,
            initial_value={"edits": []},
            args={
                "data": mo_data.csv(table_manager.to_csv()).url,
                "field-types": field_types or None,
                "pagination": pagination,
                "page-size": page_size,
            },
            on_change=on_change,
        )

    @property
    def data(
        self,
    ) -> Union[RowOrientedData, ColumnOrientedData, IntoDataFrame]:
        return self._data

    def _convert_value(
        self, value: DataEdits
    ) -> Union[RowOrientedData, ColumnOrientedData, IntoDataFrame]:
        self._edits = value
        return apply_edits(self._data, value)

    def __hash__(self) -> int:
        return id(self)


def apply_edits(
    data: Union[RowOrientedData, ColumnOrientedData, IntoDataFrame],
    edits: DataEdits,
    schema: Optional[nw.Schema] = None,
) -> Union[RowOrientedData, ColumnOrientedData, IntoDataFrame]:
    if len(edits["edits"]) == 0:
        return data
    # If row-oriented, apply edits to the data
    if isinstance(data, list):
        return _apply_edits_row_oriented(data, edits, schema)
    # If column-oriented, apply edits to the data
    elif isinstance(data, dict):
        return _apply_edits_column_oriented(data, edits, schema)

    try:
        return _apply_edits_dataframe(data, edits, schema)
    except Exception as e:
        raise ValueError(
            f"Data editor does not support this type of data: {type(data)}"
        ) from e


def _apply_edits_column_oriented(
    data: ColumnOrientedData,
    edits: DataEdits,
    schema: Optional[nw.Schema] = None,
) -> ColumnOrientedData:
    for edit in edits["edits"]:
        column = data[edit["columnId"]]
        if edit["rowIdx"] >= len(column):
            # Extend the column with None values up to the new row index
            column.extend([None] * (edit["rowIdx"] - len(column) + 1))
        dtype = schema.get(edit["columnId"]) if schema else None
        column[edit["rowIdx"]] = _convert_value(
            edit["value"], column[0] if column else None, dtype
        )

    return data


def _apply_edits_row_oriented(
    data: RowOrientedData,
    edits: DataEdits,
    schema: Optional[nw.Schema] = None,
) -> RowOrientedData:
    for edit in edits["edits"]:
        if edit["rowIdx"] >= len(data):
            # Create a new row with None values for all columns
            new_row = {col: None for col in data[0].keys()}
            data.append(new_row)
        original_value = data[0][edit["columnId"]] if data else None
        dtype = schema.get(edit["columnId"]) if schema else None
        data[edit["rowIdx"]][edit["columnId"]] = _convert_value(
            edit["value"], original_value, dtype
        )

    return data


def _apply_edits_dataframe(
    native_df: IntoDataFrame, edits: DataEdits, schema: Optional[nw.Schema]
) -> IntoDataFrame:
    df = nw.from_native(native_df, eager_or_interchange_only=True)
    column_oriented = df.to_dict(as_series=False)
    schema = schema or cast(nw.Schema, df.schema)
    new_data = _apply_edits_column_oriented(column_oriented, edits, schema)
    native_namespace = nw.get_native_namespace(df)
    new_native_df = nw.from_dict(
        new_data, native_namespace=native_namespace
    ).to_native()
    return new_native_df  # type: ignore[no-any-return]


def _convert_value(
    value: Any,
    original_value: Any,
    dtype: Optional[DType] = None,
) -> Any:
    try:
        if dtype is not None:
            if dtype == nw.Datetime:
                return datetime.datetime.fromisoformat(value)
            elif dtype == nw.Date:
                return datetime.date.fromisoformat(value)
            elif dtype == nw.Duration:
                return datetime.timedelta(microseconds=float(value))
            elif dtype == nw.Float32:
                return float(value)
            elif dtype == nw.Float64:
                return float(value)
            elif dtype == nw.Int16:
                return int(value)
            elif dtype == nw.Int32:
                return int(value)
            elif dtype == nw.Int64:
                return int(value)
            elif dtype == nw.UInt16:
                return int(value)
            elif dtype == nw.UInt32:
                return int(value)
            elif dtype == nw.UInt64:
                return int(value)
            elif dtype == nw.String:
                return str(value)
            elif dtype == nw.Enum:
                return str(value)
            elif dtype == nw.Categorical:
                return str(value)
            elif dtype == nw.Boolean:
                return bool(value)
            elif dtype == nw.List:
                # Handle list conversion
                if isinstance(value, str):
                    # Attempt to parse string as a list
                    try:
                        return list(ast.literal_eval(value))
                    except (ValueError, SyntaxError):
                        # If parsing fails, split the string
                        return value.split(",")
                elif isinstance(value, list):
                    return value  # type: ignore
                else:
                    # If it's not a string or list, wrap it in a list
                    return [value]
            else:
                LOGGER.warning(f"Unsupported dtype: {dtype}")
                return str(value)

        if original_value is None:
            return value

        # Try to convert the value to the original type
        original_type: Any = type(original_value)

        if isinstance(original_value, (int, float)):
            return original_type(value)
        elif isinstance(original_value, str):
            return str(value)
        elif isinstance(original_value, (datetime.date)):
            return datetime.date.fromisoformat(value)
        elif isinstance(original_value, (datetime.datetime)):
            return datetime.datetime.fromisoformat(value)
        elif isinstance(original_value, (datetime.timedelta)):
            return datetime.timedelta(microseconds=float(value))
        elif isinstance(original_value, list):
            # Handle list conversion
            if isinstance(value, str):
                # Attempt to parse string as a list
                try:
                    return list(ast.literal_eval(value))
                except (ValueError, SyntaxError):
                    # If parsing fails, split the string
                    return list(value.split(","))
            elif isinstance(value, list):
                return value  # type: ignore[return-value]
            else:
                # If it's not a string or list, wrap it in a list
                return [value]
        else:
            return value
    except ValueError as e:
        LOGGER.error(str(e))
        # If conversion fails, return the original value
        return original_value  # type: ignore[return-value]



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from typing import TYPE_CHECKING

from starlette.authentication import requires
from starlette.exceptions import HTTPException
from starlette.responses import HTMLResponse, PlainTextResponse

from marimo import _loggers
from marimo._server.api.deps import AppState
from marimo._server.api.status import HTTPStatus
from marimo._server.api.utils import parse_request
from marimo._server.export.exporter import AutoExporter, Exporter
from marimo._server.model import SessionMode
from marimo._server.models.export import (
    ExportAsHTMLRequest,
    ExportAsMarkdownRequest,
    ExportAsScriptRequest,
)
from marimo._server.models.models import SuccessResponse
from marimo._server.router import APIRouter

if TYPE_CHECKING:
    from starlette.requests import Request

LOGGER = _loggers.marimo_logger()

# Router for export endpoints
router = APIRouter()


@router.post("/html")
@requires("read")
async def export_as_html(
    *,
    request: Request,
) -> HTMLResponse:
    """
    requestBody:
        content:
            application/json:
                schema:
                    $ref: "#/components/schemas/ExportAsHTMLRequest"
    responses:
        200:
            description: Export the notebook as HTML
            content:
                text/html:
                    schema:
                        type: string
        400:
            description: File must be saved before downloading
    """
    app_state = AppState(request)
    body = await parse_request(request, cls=ExportAsHTMLRequest)
    session = app_state.require_current_session()

    # Only include the code and console if we are in edit mode
    if app_state.mode != SessionMode.EDIT:
        body.include_code = False

    html, filename = Exporter().export_as_html(
        file_manager=session.app_file_manager,
        session_view=session.session_view,
        display_config=app_state.session_manager.user_config_manager.get_config()[
            "display"
        ],
        request=body,
    )

    if body.download:
        headers = {"Content-Disposition": f"attachment; filename={filename}"}
    else:
        headers = {}

    # Download the HTML
    return HTMLResponse(
        content=html,
        headers=headers,
    )


@router.post("/auto_export/html")
@requires("edit")
async def auto_export_as_html(
    *,
    request: Request,
) -> SuccessResponse | PlainTextResponse:
    """
    requestBody:
        content:
            application/json:
                schema:
                    $ref: "#/components/schemas/ExportAsHTMLRequest"
    responses:
        200:
            description: Export the notebook as HTML
            content:
                application/json:
                    schema:
                        $ref: "#/components/schemas/SuccessResponse"
        400:
            description: File must be saved before downloading
    """
    app_state = AppState(request)
    body = await parse_request(request, cls=ExportAsHTMLRequest)
    session = app_state.require_current_session()
    session_view = session.session_view

    # If we have already exported to HTML, don't do it again
    if session_view.has_auto_exported_html:
        LOGGER.debug("Already auto-exported to HTML")
        return PlainTextResponse(status_code=HTTPStatus.NOT_MODIFIED)

    # Reload the file manager to get the latest state
    session.app_file_manager.reload()

    html, _filename = Exporter().export_as_html(
        file_manager=session.app_file_manager,
        session_view=session_view,
        display_config=app_state.session_manager.user_config_manager.get_config()[
            "display"
        ],
        request=body,
    )

    # Save the HTML file to disk, at `.marimo/<filename>.html`
    AutoExporter().save_html(
        file_manager=session.app_file_manager,
        html=html,
    )
    session_view.mark_auto_export_html()

    return SuccessResponse()


@router.post("/script")
@requires("edit")
async def export_as_script(
    *,
    request: Request,
) -> PlainTextResponse:
    """
    requestBody:
        content:
            application/json:
                schema:
                    $ref: "#/components/schemas/ExportAsScriptRequest"
    responses:
        200:
            description: Export the notebook as a script
            content:
                text/plain:
                    schema:
                        type: string
        400:
            description: File must be saved before downloading
    """
    app_state = AppState(request)
    body = await parse_request(request, cls=ExportAsScriptRequest)
    session = app_state.require_current_session()

    python, filename = Exporter().export_as_script(
        file_manager=session.app_file_manager,
    )

    if body.download:
        headers = {"Content-Disposition": f"attachment; filename={filename}"}
    else:
        headers = {}

    # Download the Script
    return PlainTextResponse(
        content=python,
        headers=headers,
    )


@router.post("/markdown")
@requires("edit")
async def export_as_markdown(
    *,
    request: Request,
) -> PlainTextResponse:
    """
    requestBody:
        content:
            application/json:
                schema:
                    $ref: "#/components/schemas/ExportAsMarkdownRequest"
    responses:
        200:
            description: Export the notebook as a markdown
            content:
                text/plain:
                    schema:
                        type: string
        400:
            description: File must be saved before downloading
    """
    app_state = AppState(request)
    body = await parse_request(request, cls=ExportAsMarkdownRequest)
    app_file_manager = app_state.require_current_session().app_file_manager
    # Reload the file manager to get the latest state
    app_file_manager.reload()

    if not app_file_manager.path:
        raise HTTPException(
            status_code=HTTPStatus.BAD_REQUEST,
            detail="File must be saved before downloading",
        )

    markdown, filename = Exporter().export_as_md(
        file_manager=app_file_manager,
    )

    if body.download:
        headers = {"Content-Disposition": f"attachment; filename={filename}"}
    else:
        headers = {}

    # Download the Markdown
    return PlainTextResponse(
        content=markdown,
        headers=headers,
    )


@router.post("/auto_export/markdown")
@requires("edit")
async def auto_export_as_markdown(
    *,
    request: Request,
) -> SuccessResponse | PlainTextResponse:
    """
    requestBody:
        content:
            application/json:
                schema:
                    $ref: "#/components/schemas/ExportAsMarkdownRequest"
    responses:
        200:
            description: Export the notebook as a markdown
            content:
                application/json:
                    schema:
                        $ref: "#/components/schemas/SuccessResponse"
        400:
            description: File must be saved before downloading
    """
    app_state = AppState(request)
    session = app_state.require_current_session()
    session_view = session.session_view

    # If we have already exported to Markdown, don't do it again
    if session_view.has_auto_exported_md:
        LOGGER.debug("Already auto-exported to Markdown")
        return PlainTextResponse(status_code=HTTPStatus.NOT_MODIFIED)

    # Reload the file manager to get the latest state
    session.app_file_manager.reload()

    markdown, _filename = Exporter().export_as_md(
        file_manager=session.app_file_manager,
    )

    # Save the Markdown file to disk, at `.marimo/<filename>.md`
    AutoExporter().save_md(
        file_manager=session.app_file_manager,
        markdown=markdown,
    )
    session_view.mark_auto_export_md()

    return SuccessResponse()



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

import abc
from enum import Enum
from typing import TYPE_CHECKING, Callable

from marimo._server.ids import ConsumerId

if TYPE_CHECKING:
    from marimo._messaging.ops import MessageOperation
    from marimo._messaging.types import KernelMessage


class ConnectionState(Enum):
    """Connection state for a session"""

    OPEN = 0
    CLOSED = 1
    ORPHANED = 2


class SessionMode(str, Enum):
    """Session mode for a session"""

    # read-write
    EDIT = "edit"
    # read-only
    RUN = "run"


class SessionConsumer(metaclass=abc.ABCMeta):
    """
    Consumer for a session

    This allows use to communicate with a session via different
    connection types. Currently we consume a session via WebSocket
    """

    def __init__(self, consumer_id: ConsumerId) -> None:
        self.consumer_id = consumer_id

    @abc.abstractmethod
    def on_start(
        self,
    ) -> Callable[[KernelMessage], None]:
        """
        Start the session consumer
        and return a subscription function for the session consumer
        """
        raise NotImplementedError

    @abc.abstractmethod
    def on_stop(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def write_operation(self, op: MessageOperation) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def connection_state(self) -> ConnectionState:
        raise NotImplementedError



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Optional, Union

from marimo._ast.cell import CellId_t
from marimo._runtime.dataflow import EdgeWithVar


@dataclass
class CycleError:
    edges_with_vars: tuple[EdgeWithVar, ...]
    type: Literal["cycle"] = "cycle"


@dataclass
class MultipleDefinitionError:
    name: str
    cells: tuple[CellId_t, ...]
    type: Literal["multiple-defs"] = "multiple-defs"


@dataclass
class DeleteNonlocalError:
    name: str
    cells: tuple[CellId_t, ...]
    type: Literal["delete-nonlocal"] = "delete-nonlocal"


@dataclass
class MarimoInterruptionError:
    type: Literal["interruption"] = "interruption"


@dataclass
class MarimoAncestorPreventedError:
    msg: str
    raising_cell: CellId_t
    blamed_cell: Optional[CellId_t]
    type: Literal["ancestor-prevented"] = "ancestor-prevented"


@dataclass
class MarimoAncestorStoppedError:
    msg: str
    raising_cell: CellId_t
    type: Literal["ancestor-stopped"] = "ancestor-stopped"


@dataclass
class MarimoExceptionRaisedError:
    msg: str
    exception_type: str
    # None for if raising_cell is the current cell
    raising_cell: Optional[CellId_t]
    type: Literal["exception"] = "exception"


@dataclass
class MarimoSyntaxError:
    msg: str
    type: Literal["syntax"] = "syntax"


@dataclass
class UnknownError:
    msg: str
    type: Literal["unknown"] = "unknown"


@dataclass
class MarimoStrictExecutionError:
    msg: str
    ref: str
    blamed_cell: Optional[CellId_t]
    type: Literal["strict-exception"] = "strict-exception"


Error = Union[
    CycleError,
    MultipleDefinitionError,
    DeleteNonlocalError,
    MarimoAncestorStoppedError,
    MarimoAncestorPreventedError,
    MarimoExceptionRaisedError,
    MarimoStrictExecutionError,
    MarimoInterruptionError,
    MarimoSyntaxError,
    UnknownError,
]



# Copyright 2024 Marimo. All rights reserved.

import marimo

__generated_with = "0.9.2"
app = marimo.App()


@app.cell
def __():
    import marimo as mo

    mo.md("# Welcome to marimo! 🌊🍃")
    return (mo,)


@app.cell
def __(mo):
    slider = mo.ui.slider(1, 22)
    return (slider,)


@app.cell
def __(mo, slider):
    mo.md(
        f"""
        marimo is a **reactive** Python notebook.

        This means that unlike traditional notebooks, marimo notebooks **run
        automatically** when you modify them or
        interact with UI elements, like this slider: {slider}.

        {"##" + "🍃" * slider.value}
        """
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.accordion(
        {
            "Tip: disabling automatic execution": mo.md(
                rf"""
            marimo lets you disable automatic execution: just go into the
            notebook settings and set

            "Runtime > On Cell Change" to "lazy".

            When the runtime is lazy, after running a cell, marimo marks its
            descendants as stale instead of automatically running them. The
            lazy runtime puts you in control over when cells are run, while
            still giving guarantees about the notebook state.
            """
            )
        }
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        Tip: This is a tutorial notebook. You can create your own notebooks
        by entering `marimo edit` at the command line.
        """
    ).callout()
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        ## 1. Reactive execution

        A marimo notebook is made up of small blocks of Python code called
        cells.

        marimo reads your cells and models the dependencies among them: whenever
        a cell that defines a global variable  is run, marimo
        **automatically runs** all cells that reference that variable.

        Reactivity keeps your program state and outputs in sync with your code,
        making for a dynamic programming environment that prevents bugs before they
        happen.
        """
    )
    return


@app.cell(hide_code=True)
def __(changed, mo):
    (
        mo.md(
            f"""
            **✨ Nice!** The value of `changed` is now {changed}.

            When you updated the value of the variable `changed`, marimo
            **reacted** by running this cell automatically, because this cell
            references the global variable `changed`.

            Reactivity ensures that your notebook state is always
            consistent, which is crucial for doing good science; it's also what
            enables marimo notebooks to double as tools and  apps.
            """
        )
        if changed
        else mo.md(
            """
            **🌊 See it in action.** In the next cell, change the value of the
            variable  `changed` to `True`, then click the run button.
            """
        )
    )
    return


@app.cell
def __():
    changed = False
    return (changed,)


@app.cell(hide_code=True)
def __(mo):
    mo.accordion(
        {
            "Tip: execution order": (
                """
                The order of cells on the page has no bearing on
                the order in which cells are executed: marimo knows that a cell
                reading a variable must run after the cell that  defines it. This
                frees you to organize your code in the way that makes the most
                sense for you.
                """
            )
        }
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        **Global names must be unique.** To enable reactivity, marimo imposes a
        constraint on how names appear in cells: no two cells may define the same
        variable.
        """
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.accordion(
        {
            "Tip: encapsulation": (
                """
                By encapsulating logic in functions, classes, or Python modules,
                you can minimize the number of global variables in your notebook.
                """
            )
        }
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.accordion(
        {
            "Tip: private variables": (
                """
                Variables prefixed with an underscore are "private" to a cell, so
                they can be defined by multiple cells.
                """
            )
        }
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        ## 2. UI elements

        Cells can output interactive UI elements. Interacting with a UI
        element **automatically triggers notebook execution**: when
        you interact with a UI element, its value is sent back to Python, and
        every cell that references that element is re-run.

        marimo provides a library of UI elements to choose from under
        `marimo.ui`.
        """
    )
    return


@app.cell
def __(mo):
    mo.md("""**🌊 Some UI elements.** Try interacting with the below elements.""")
    return


@app.cell
def __(mo):
    icon = mo.ui.dropdown(["🍃", "🌊", "✨"], value="🍃")
    return (icon,)


@app.cell
def __(icon, mo):
    repetitions = mo.ui.slider(1, 16, label=f"number of {icon.value}: ")
    return (repetitions,)


@app.cell
def __(icon, repetitions):
    icon, repetitions
    return


@app.cell
def __(icon, mo, repetitions):
    mo.md("# " + icon.value * repetitions.value)
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        ## 3. marimo is just Python

        marimo cells parse Python (and only Python), and marimo notebooks are
        stored as pure Python files — outputs are _not_ included. There's no
        magical syntax.

        The Python files generated by marimo are:

        - easily versioned with git, yielding minimal diffs
        - legible for both humans and machines
        - formattable using your tool of choice,
        - usable as Python  scripts, with UI  elements taking their default
        values, and
        - importable by other modules (more on that in the future).
        """
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        ## 4. Running notebooks as apps

        marimo notebooks can double as apps. Click the app window icon in the
        bottom-right to see this notebook in "app view."

        Serve a notebook as an app with `marimo run` at the command-line.
        Of course, you can use marimo just to level-up your
        notebooking, without ever making apps.
        """
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        ## 5. The `marimo` command-line tool

        **Creating and editing notebooks.** Use

        ```
        marimo edit
        ```

        in a terminal to start the marimo notebook server. From here
        you can create a new notebook or edit existing ones.


        **Running as apps.** Use

        ```
        marimo run notebook.py
        ```

        to start a webserver that serves your notebook as an app in read-only mode,
        with code cells hidden.

        **Convert a Jupyter notebook.** Convert a Jupyter notebook to a marimo
        notebook using `marimo convert`:

        ```
        marimo convert your_notebook.ipynb > your_app.py
        ```

        **Tutorials.** marimo comes packaged with tutorials:

        - `dataflow`: more on marimo's automatic execution
        - `ui`: how to use UI elements
        - `markdown`: how to write markdown, with interpolated values and
           LaTeX
        - `plots`: how plotting works in marimo
        - `sql`: how to use SQL
        - `layout`: layout elements in marimo
        - `fileformat`: how marimo's file format works
        - `markdown-format`: for using `.md` files in marimo
        - `for-jupyter-users`: if you are coming from Jupyter

        Start a tutorial with `marimo tutorial`; for example,

        ```
        marimo tutorial dataflow
        ```

        In addition to tutorials, we have examples in our
        [our GitHub repo](https://www.github.com/marimo-team/marimo/tree/main/examples).
        """
    )
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        ## 6. The marimo editor

        Here are some tips to help you get started with the marimo editor.
        """
    )
    return


@app.cell
def __(mo, tips):
    mo.accordion(tips)
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md("""## Finally, a fun fact""")
    return


@app.cell(hide_code=True)
def __(mo):
    mo.md(
        """
        The name "marimo" is a reference to a type of algae that, under
        the right conditions, clumps together to form a small sphere
        called a "marimo moss ball". Made of just strands of algae, these
        beloved assemblages are greater than the sum of their parts.
        """
    )
    return


@app.cell(hide_code=True)
def __():
    tips = {
        "Saving": (
            """
            **Saving**

            - _Name_ your app using the box at the top of the screen, or
              with `Ctrl/Cmd+s`. You can also create a named app at the
              command line, e.g., `marimo edit app_name.py`.

            - _Save_ by clicking the save icon on the bottom right, or by
              inputting `Ctrl/Cmd+s`. By default marimo is configured
              to autosave.
            """
        ),
        "Running": (
            """
            1. _Run a cell_ by clicking the play ( ▷ ) button on the top
            right of a cell, or by inputting `Ctrl/Cmd+Enter`.

            2. _Run a stale cell_  by clicking the yellow run button on the
            right of the cell, or by inputting `Ctrl/Cmd+Enter`. A cell is
            stale when its code has been modified but not run.

            3. _Run all stale cells_ by clicking the play ( ▷ ) button on
            the bottom right of the screen, or input `Ctrl/Cmd+Shift+r`.
            """
        ),
        "Console Output": (
            """
            Console output (e.g., `print()` statements) is shown below a
            cell.
            """
        ),
        "Creating, Moving, and Deleting Cells": (
            """
            1. _Create_ a new cell above or below a given one by clicking
                the plus button to the left of the cell, which appears on
                mouse hover.

            2. _Move_ a cell up or down by dragging on the handle to the 
                right of the cell, which appears on mouse hover.

            3. _Delete_ a cell by clicking the trash bin icon. Bring it
                back by clicking the undo button on the bottom right of the
                screen, or with `Ctrl/Cmd+Shift+z`.
            """
        ),
        "Disabling Automatic Execution": (
            """
            Via the notebook settings (gear icon) or footer panel, you
            can disable automatic execution. This is helpful when
            working with expensive notebooks or notebooks that have
            side-effects like database transactions.
            """
        ),
        "Disabling Cells": (
            """
            You can disable a cell via the cell context menu.
            marimo will never run a disabled cell or any cells that depend on it.
            This can help prevent accidental execution of expensive computations
            when editing a notebook.
            """
        ),
        "Code Folding": (
            """
            You can collapse or fold the code in a cell by clicking the arrow
            icons in the line number column to the left, or by using keyboard
            shortcuts.

            Use the command palette (`Ctrl/Cmd+k`) or a keyboard shortcut to
            quickly fold or unfold all cells.
            """
        ),
        "Code Formatting": (
            """
            If you have [ruff](https://github.com/astral-sh/ruff) installed,
            you can format a cell with the keyboard shortcut `Ctrl/Cmd+b`.
            """
        ),
        "Command Palette": (
            """
            Use `Ctrl/Cmd+k` to open the command palette.
            """
        ),
        "Keyboard Shortcuts": (
            """
            Open the notebook menu (top-right) or input `Ctrl/Cmd+Shift+h` to
            view a list of all keyboard shortcuts.
            """
        ),
        "Configuration": (
            """
           Configure the editor by clicking the gears icon near the top-right
           of the screen.
           """
        ),
    }
    return (tips,)


if __name__ == "__main__":
    app.run()



# Copyright 2024 Marimo. All rights reserved.
"""AI utilities."""

from __future__ import annotations

__all__ = [
    "ChatMessage",
    "ChatModelConfig",
    "ChatAttachment",
    "llm",
]

from marimo._ai import llm
from marimo._ai.types import (
    ChatAttachment,
    ChatMessage,
    ChatModelConfig,
)



# Copyright 2024 Marimo. All rights reserved.
from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from multiprocessing.connection import Connection

T = TypeVar("T")


class TypedConnection(Generic[T]):
    """Wrapper around a connection with strong typing."""

    def __init__(self, delegate: Connection):
        self._delegate = delegate

    @classmethod
    def of(
        cls,
        delegate: Connection,
    ) -> TypedConnection[T]:
        """Create a typed connection from a connection."""
        return delegate  # type: ignore[return-value]

    def send(self, obj: T) -> None:
        self._delegate.send(obj)

    def recv(self) -> T:
        return self._delegate.recv()  # type: ignore[no-any-return]

    def poll(self) -> bool:
        return self._delegate.poll()

    def fileno(self) -> int:
        return self._delegate.fileno()

    @property
    def closed(self) -> bool:
        return self._delegate.closed

    def close(self) -> None:
        self._delegate.close()



# Copyright 2024 Marimo. All rights reserved.
import functools
import sys


@functools.wraps(input)
def input_override(prompt: str = "") -> str:
    # sys.stdin is overridden
    return sys.stdin._readline_with_prompt(prompt)  # type: ignore[attr-defined, no-any-return]  # noqa: E501
